# -*- coding: utf-8 -*-
from __future__ import print_function, division
from PyAstronomy import funcFit as _fuf
from PyAstronomy.funcFit import _pymcImport, _scoImport, ic
import numpy as np
import re
import copy
import types
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyaC
import six
import six.moves as smo
import bidict
import collections
import sys
import scipy.optimize as sco
import inspect
import sys
import fnmatch
from PyAstronomy import modelSuite as _ms

_fbo2module = sys.modules[__name__]


if ic.check["progressbar"]:
    import progressbar
if ic.check["emcee"]:
    import emcee


class PyAPa(object):

    def setVal(self, v):
        self._value = v

    def getVal(self):
        return self._value

    def setFree(self, v):
        if (v == True) and (not self.thawable):
            raise (PE.PyAValError("Cannot thaw requested parameter.",
                                  solution="Some parameters cannot be thawed. See the documentation for details."))
        self._free = v

    def getFree(self):
        return self._free

    value = property(getVal, setVal)
    free = property(getFree, setFree)

    def __init__(self, value=0.0):
        self._value = value
        self._free = False
        self.affects = set()
        self.relation = None
        self.dependsOn = None
        self.thawable = True
        self.restriction = (None, None)

    def restrict(self, lower=None, upper=None):
        lin = not lower is None
        uin = not upper is None
        if lin+uin == 0:
            # No limits defined (or removed)
            pass
        elif lin+uin == 2:
            if lower >= upper:
                raise (PE.PyAValError(
                    "The lower limit has to be smaller than the upper bound in restriction."))
        self.restriction = (lower, upper)

    def isRestricted(self):
        return any([(not r is None) for r in self.restriction])

    def isRelated(self):
        return (not self.relation is None)

    def updateRelation(self):
        r"""
        Update value according to relation
        """
        if self.isRelated():
            self._value = self.relation.update()


class PyARelation(object):

    def __init__(self, ivs, func):
        r"""
        Parameters
        ----------
        ivs : list of Parameter instances
            The independent parameters
        func : Callable
            The functional relation
        """
        self.ivs = ivs
        self.func = func

    def update(self):
        return self.func(*[p.value for p in self.ivs])


class PyAPrior(object):

    def __init__(self, ivs, func, descr=None):
        r"""
        Parameters
        ----------
        ivs : list of Parameter instances
            The independent parameters
        func : Callable
            Function returning log(prior) depending on the value of the
            independent parameters
        descr : string, optional
            Description of the prior
        """
        if not hasattr(ivs, "__iter__"):
            ivs = [ivs]
        self.ivs = ivs
        self.func = func
        if descr is None:
            self.descr = "Pya Prior"
        else:
            self.descr = descr

    def evaluate(self):
        return self.func(*[p.value for p in self.ivs])

    def __str__(self):
        return str(self.descr)


class PyAUniformPrior(PyAPrior):

    def __init__(self, ivs, lower=None, upper=None, descr=""):

        if (lower is None) and (upper is None):
            raise (PE.PyAValError(
                "At least one of lower and upper must be specified"))

        if (lower is None) and (upper is None):
            def f(x): return 0.0
        elif (lower is None) and (not upper is None):
            def f(x): return 0.0 if x <= upper else -np.inf
        elif (not lower is None) and (upper is None):
            def f(x): return 0.0 if x >= lower else -np.inf
        else:
            if upper <= lower:
                raise (PE.PyAValError("'lower' must be smaller than 'upper'",
                                      where="addUniformPrior"))
            r = upper - lower

            def f(x):
                y = np.log(r) if (x >= lower) and (x <= upper) else -np.inf
                return y

        PyAPrior.__init__(self, ivs, f, descr=descr)


class PyANormalPrior(PyAPrior):

    def __init__(self, ivs, mean=None, std=None, descr=""):

        if (mean is None) and (std is None):
            raise (PE.PyAValError("Both mean and std must be specified"))

        const = -0.5*np.log(2*np.pi*std**2)
        durch = 2*std**2

        def f(x): return const - (x-mean)**2/durch

        PyAPrior.__init__(self, ivs, f, descr=descr)


class PyASmoothUniformPrior(PyAPrior):

    def __init__(self, ivs, lower=None, upper=None, scale=1e-9, descr=""):

        if (lower is None) and (upper is None):
            raise (PE.PyAValError(
                "At least one of lower and upper must be specified"))

        pih = np.pi/2.0
        flmax = sys.float_info.max/1e3

        def f(x):
            if not lower is None:
                if x < lower:
                    r = -np.arctan((lower-x)/scale)/pih*flmax
                    return r
            if not upper is None:
                if x > upper:
                    r = -np.arctan((x-upper)/scale)/pih*flmax
                    return r
            return 0.0

        PyAPrior.__init__(self, ivs, f, descr=descr)


class PyAScalePrior(PyAPrior):

    def __init__(self, ivs, m=-1., descr=""):

        def f(x):
            if x <= 0.0:
                return -np.inf
            return m*np.log(x)

        PyAPrior.__init__(self, ivs, f, descr=descr)


class PyACauchyPrior(PyAPrior):

    def __init__(self, ivs, scale, x0=0., side="both", descr=""):

        def f(x):

            pdf = np.log(scale) - np.log(np.pi) - np.log(scale**2 + (x-x0)**2)
            if side == "both":
                return pdf
            elif side == "positive":
                if (x < x0):
                    return -np.inf
                # An extra factor accounts for the 'missing' other side
                return np.log(2.) + pdf
            elif side == "negative":
                if (x > x0):
                    return -np.inf
                # An extra factor accounts for the 'missing' other side
                return np.log(2.) + pdf
            else:
                raise (PE.PyAValError("Unknown side parameter in PyACauchyPrior (" + str(side) + ")",
                                      solution="Use either of 'both', 'positive', 'negative'"))

        PyAPrior.__init__(self, ivs, f, descr=descr)


class PyABPS(object):
    r"""
    Basic Parameter Set
    """

    def addParam(self, name, value=0.0):
        r"""
        Add parameter to the list

        Parameters
        ----------
        name : string
            Name of the parameter
        value : optional
            Default value of the parameter
        """
        if name in self.pmap:
            raise (PE.PyAValError("Name '" + str(name) + "' already exists!",
                                  where="addPyAPS::Param"))
        self.pmap[name] = PyAPa(value)

    def _checkParam(self, n):
        """ Throws and exception if parameter 'n' does not exist """
        if not n in self.pmap:
            raise (PE.PyAValError("No such parameter: " + str(n),
                                  solution="Use one of: " + ', '.join(list(self.pmap))))

    def __getitem__(self, n, ref=False):
        """ Get reference to PyaPa """
        self._checkParam(n)
        if ref:
            return self.pmap[n]
        else:
            return self.pmap[n].value

    def __setitem__(self, n, v):
        """ Set value and update related if necessary """
        self._checkParam(n)
        self.pmap[n].value = v
        if len(self.pmap[n].affects) > 0:
            for a in self.pmap[n].affects:
                a.updateRelation()

    def copy(self):
        """ Return a shallow copy of the object """
        c = PyABPS([], self.rootName, number=self.number)
        c.pmap = copy.copy(self.pmap)
        return c

    def __init__(self, pns, rootName, number=0):
        self.rootName = rootName
        self.number = number
        self.pmap = bidict.OrderedBidict()
        for n in pns:
            self.addParam(n, 0.0)


class PyABaSoS(object):
    r"""
    Basic Set of Basic Parameter Sets
    """

    def _allRoots(self):
        """ Get root names of all basic sets """
        return [s.rootName for s in self.bss]

    def _rootNumbers(self, rn):
        """ All number associated with a certain root """
        return set([b.number if b.rootName == rn else 0 for b in self.bss])

    def _updatepmap(self):
        """ Update parameter names according to root name and number. """
        ar = self._allRoots()
        if len(ar) == 1:
            # Only one subset. Apply no name updating
            self.pmap = bidict.OrderedBidict()
            for n in list(self.bss[0].pmap):
                self.pmap[n] = self.bss[0].pmap[n]
        else:
            # More than one subset
            # Root Count
            rc = {r: ar.count(r) for r in set(ar)}
            self.pmap = bidict.OrderedBidict()
            for b in self.bss:
                useNumber = int(rc[b.rootName] > 1)
                for n in list(b.pmap):
                    useName = self._composePN(
                        n, b.rootName, b.number*useNumber)
                    self.pmap[useName] = b.pmap[n]

    def addBPS(self, s):
        r"""
        Add Basic Parameter Set to set of sets

        Parameters
        ----------
        s : PyABPS
            Instance of PyABPS
        """
        snew = s.copy()
        rns = self._rootNumbers(snew.rootName)
        if len(rns) == 0:
            snew.number = 1
        else:
            snew.number = max(rns)+1
        self.bss.append(snew)
        self._updatepmap()

    def _checkParam(self, n):
        """ Throws and exception if parameter 'n' does not exist """
        if not n in self.pmap:
            raise (PE.PyAValError("No such parameter: " + str(n),
                                  solution="Use one of: " + ', '.join(list(self.pmap))))

    def _lazyNameMatch(self, n):
        r"""
        Search by 'part of name'

        Searches the parameter map for occurrences of the specified
        sub-string, tests whether it is unique, and returns the full
        valid name if so. 

        Parameters
        ----------
        n : string
            (Unique) Part of the parameter name

        Returns
        -------
        True name : string
        """
        try:
            self._checkParam(n)
            return n
        except PE.PyAValError as e:
            matches = [pn for pn in list(self.pmap) if pn.find(n) != -1]
            if len(matches) == 1:
                return matches[0]
            elif len(matches) > 1:
                raise (PE.PyAValError("More than one match in lazy name matching for parameter '" +
                       str(n) + "'. Matches: " + ', '.join(matches)))
            else:
                raise (PE.PyAValError(
                    "No match in lazy name matching for parameter '" + str(n) + "'."))

    def __getitem__(self, n):
        """ Get value of parameter """
        n = self._lazyNameMatch(n)
        return self.pmap[n].value

    def getPRef(self, n):
        """ Get reference to PyaPa """
        self._checkParam(n)
        return self.pmap[n]

    def __setitem__(self, n, v):
        """ Set value and update related if necessary """
        n = self._lazyNameMatch(n)
        self.pmap[n].value = v
        if len(self.pmap[n].affects) > 0:
            for a in self.pmap[n].affects:
                a.updateRelation()

    def rename(self, old, new):
        r"""
        Rename parameter

        Parameters
        ----------
        old, new : strings
            Old and new name
        """
        self._checkParam(old)
        if new in self.pmap:
            raise (PE.PyAValError("Parameter " + str(new) + " already exists.",
                                  solution="Use a unique name."))
        tmp = self.pmap[old]
        del self.pmap[old]
        self.pmap[new] = tmp

    def relate(self, dv, idv, func=None):
        r"""
        Establish functional relations

        Parameters
        ----------
        dv : string
            Dependent variable
        idv : string or list of strings
            Independent variables
        func : callable
            Callable taking the independent variables as arguments
            and returns the value of the dependent variable.
        """
        if isinstance(idv, six.string_types):
            # Convert single string into list
            idv = [idv]
        # Check all parameters
        self._checkParam(dv)
        [self._checkParam(n) for n in idv]

        # By default, use equal
        if func is None:
            def func(x): return x

        # Manage dependent variable
        self.pmap[dv].dependsOn = [self.pmap[n] for n in idv]
        self.pmap[dv].relation = PyARelation(self.pmap[dv].dependsOn, func)
        # Manage affected variables
        for n in idv:
            self.pmap[n].affects.update([self.pmap[dv]])
        # Trigger update of the value of the related parameter
        self.pmap[dv].updateRelation()
        # Freeze the dependent variable
        self.freeze(dv)

    def copy(self):
        """ Return a shallow copy of the object """
        c = PyABaSoS()
        c.pmap = copy.copy(self.pmap)
        c.bss = copy.copy(self.bss)
        return c

    @classmethod
    def combine(cls, left, right):
        r"""
        Combine parameter sets

        Parameters
        ----------
        left, right : PyABaSoS
            Instances of PyABaSoS holding the individual sets

        Returns
        -------
        Combined object : PyABaSoS
        """
        c = cls()
        c.bss = []
        for b in left.bss:
            c.addBPS(b)
        for b in right.bss:
            c.addBPS(b)
        return c

    def _composePN(self, b, r, n):
        r"""
        Compose parameter name

        Parameters
        ----------
        b, r : strings
            Base and root
        n : int
            Number

        Returns
        -------
        Name : string
            Complete parameter name
        """
        result = b
        if (not r is None) and (r != ""):
            result += "_" + r
        if (not n is None) and (n != 0):
            result += ("(%d)" % n)
        return result

    def _decomposePN(self, n):
        r"""
        Decompose parameter name into Base, Root, Number
        """
        r = re.match(r"([^_]+)((_([^\(]+)?)?(\(([0-9]+)\))?)?", n)
        if r is None:
            return None, None, None
        else:
            return r.group(0), r.group(3), r.group(5)

    def parameters(self):
        r"""
        Get all parameter names and values

        Returns
        -------
        Parameters : OrderedDict (collections)
            (Ordered) Dictionary mapping parameter name to value
        """
        ps = collections.OrderedDict()
        for k, v in six.iteritems(self.pmap):
            ps[k] = v.value
        return ps

    def freeParamNames(self):
        """ Get list of names of free parameters """
        return [p for p in list(self.pmap) if self.pmap[p].free]

    def freeParameters(self):
        r"""
        Get a dictionary mapping parameter name to value
        """
        fpn = self.freeParamNames()
        fpv = self.freeParamVals()
        return dict(zip(fpn, fpv))

    def getRestrictions(self):
        """ Get dictionary mapping parameter names to restrictions """
        return {p: self.pmap[p].restriction for p in list(self.pmap) if self.pmap[p].isRestricted()}

    def setRestriction(self, restrictions):
        r"""
        Add restrictions for parameters

        Parameters
        ----------
        restrictions : dictionary
            Maps parameter to restriction. A restriction is a two-tuple
            with lower and upper limit. If None is specified, no restriction
            applies in this direction.
        """
        for k, v in six.iteritems(restrictions):
            self._checkParam(k)
            self.getPRef(k).restrict(lower=v[0], upper=v[1])

    def freeParamVals(self):
        """ Get list of names of free parameters """
        return [self.pmap[p].value for p in list(self.pmap) if self.pmap[p].free]

    def setFreeParamVals(self, vals):
        r"""
        Assign values to free parameters.

        The order is fixed (use, e.g., freeParamNames to get it).

        Parameters
        ----------
        vals : list or array
            The values to be assigned
        """
        try:
            # Handle 'singleton' arrays
            if len(vals.shape) == 0:
                vals = np.array([vals])
        except:
            pass
        for i, p in enumerate(self.freeParamNames()):
            self[p] = vals[i]

    def _thawfreeze(self, pns, free):
        r"""
        Thaw or freeze parameters

        Parameters
        ----------
        pns : string or list of strings
            Relevant parameters
        free : boolean
            True to thaw and False to freeze
        """
        if isinstance(pns, six.string_types):
            pns = [pns]
        for p in pns:
            self._checkParam(p)
            self.pmap[p].free = free

    def thaw(self, pns):
        r"""
        Thaw parameters

        Parameters
        ----------
        pns : string or list of strings
            Relevant parameters
        """
        self._thawfreeze(pns, True)

    def freeze(self, pns):
        r"""
        Freeze parameters

        Parameters
        ----------
        pns : string or list of strings
            Relevant parameters
        """
        self._thawfreeze(pns, False)

    def __init__(self, *args):
        self.bss = []
        for s in args:
            self.addBPS(s)
        self._updatepmap()


def gaussLogL(x, y, yerr, m):
    r"""
    Gaussian likelihood

    Parameters
    ----------
    x, y : arrays
        The x and y coordinates of the data points.
    yerr : array or float
        Uncertainty of data points
    m : array
        The model

    Returns
    -------
    lnl : float
        The natural logarithm of the likelihood.
    """
    if hasattr(yerr, "__iter__"):
        # yerr is an array
        lnl = -len(x)/2.0*np.log(2.*np.pi) - np.sum(np.log(yerr)) - \
            0.5 * np.sum((m-y)**2/(yerr**2))
        return lnl
    else:
        # yerr is a float
        lnl = -len(x)/2.0*np.log(2.*np.pi) - len(x) * \
            np.log(yerr) - 0.5 * np.sum((m-y)**2/(yerr**2))
        return lnl


def _gaussLogL(self, *args, **kwargs):
    r"""
    Default implementation of Gaussian likelihood.

    Parameters
    ----------
    x, y : arrays
        The x and y coordinates of the data points.
    yerr : array or float, optional
        If not specified, a value of 1.0 will be assumed. Otherwise interpreted as
        the error of the data points.

    Returns
    -------
    lnl : float
        The natural logarithm of the likelihood based on a model f(x), data points (x,y), and
        normal errors, yerr.
    """
    if len(args) == 2:
        x, y = args[0], args[1]
        yerr = 1.0
    elif len(args) >= 3:
        x, y, yerr = args[0], args[1], args[2]
    else:
        raise (PE.PyAValError("Invalid call to _gaussLogL. Received " +
               str(len(args)) + " arguments but takes 2 or 3 (x, y, [yerr])."))

    if "_currentModel" in kwargs:
        m = kwargs["_currentModel"]
    else:
        m = self.evaluate(*args, **kwargs)

    return gaussLogL(x, y, yerr, m)


def getchisqrobjf(sumfct=np.sum):

    def chisqrobjf(self, pars, *args, **kwargs):
        r"""
        Default implementation of chi-square objective function

        Parameters
        ----------
        x, y : arrays
            The x and y coordinates of the data points.
        yerr : array or float, optional
            If not specified, a value of 1.0 will be assumed. Otherwise interpreted as
            the error of the data points.

        Returns
        -------
        chi square : float
            Returns chi square if uncertainty is specified and sum of squared residuals
            otherwise.
        """
        if len(args) == 2:
            x, y = args[0], args[1]
            yerr = 1.0
        elif len(args) >= 3:
            x, y, yerr = args[0], args[1], args[2]
        else:
            raise (PE.PyAValError("Invalid call to _chisqr. Received " +
                   str(len(args)) + " arguments but takes 2 or 3 (x, y, [yerr])."))

        if "_currentModel" in kwargs:
            m = kwargs["_currentModel"]
        else:
            m = self.evaluate(*args, **kwargs)

        return sumfct((m-y)**2/yerr**2)
    return chisqrobjf


class MBO(object):
    r"""
    Model Base Object

    Concepts
    --------

    Parameter names: Internal vs. external



    """

    def __init__(self, pars=None, rootName="", **kwargs):

        if pars is None:
            raise (PE.PyAValError("You need to specify the parameter names via 'pars'.",
                                  where="MBO2",
                                  solution="Specify something along the lines of 'pars = ['pn1', 'pn2', ...]'."))

        self.pars = (PyABaSoS(PyABPS(pars, rootName)))
        self.priors = []

        # Penalty factor used in objective function
        self._pf = 1e20

        self.rootName = rootName
        self.leftCompo = None
        self.rightCompo = None

    def _combineMBOs(self, right):
        r"""
        Combine two MBO2s into a new one

        Parameters
        ----------
        right : MBO2 instance
            MBO object on the right hand side of the operation

        Returns
        -------
        combination : MBO2
        """
        r = MBO([], rootName="combined")
        r.pars = PyABaSoS.combine(self.pars, right.pars)
        r.leftCompo = self
        r.rightCompo = right
        return r

    def __add__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) + self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def __sub__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) - self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def __mul__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) * self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def __div__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) / self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def __truediv__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) / self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def __pow__(self, right):
        result = self._combineMBOs(right)
        result.evaluate = types.MethodType(lambda self, *args, **kwargs:
                                           self.leftCompo.evaluate(*args, **kwargs) ** self.rightCompo.evaluate(*args, **kwargs), result)
        return result

    def parameterSummary(self, toScreen=True, prefix=""):
        r"""
        Parameter summary

        Parameters
        ----------
        toScreen : boolean, optional
            If True (default), the parameter summary is printed to screen.
        prefix : string, optional
            If specified, the prefix is added at the front of all lines.

        Returns
        -------
        Summary : list of lines
            The summary as a list of lines
        """
        # Max length of parameter name
        mpl = max([len(n) for n in list(self.parameters())])
        lines = []
        tf = {True: "T", False: "F"}
        for k, v in self.pars.pmap.items():
            l = prefix + "    " + \
                ("%" + str(mpl) + "s") % k + " = % 12g" % v.value
            l += ", free: " + tf[v.free] + \
                ", restricted: " + tf[v.isRestricted()]
            l += ", related: " + tf[v.isRelated()]
            lines.append(l)
#         lines.append(prefix + "Priors: ")
#         if len(self.priors) > 0:
#             for i, p in enumerate(self.priors, 1):
#                 lines.append(prefix + " "*4 + ("%2d) " % i) + str(p))
#         else:
#             lines.append("    All uniform (=1)")

        if toScreen:
            mll = max([len(l) for l in lines])
            print("-" * (mll//3) + " Parameter summary " + "-"*(mll-mll//3-19))
            for l in lines:
                print(l)
            print("-" * mll)
        return lines

    def relate(self, dv, idv, func=None):
        self.pars.relate(dv, idv, func)

    def __getitem__(self, n):
        return self.pars[n]

    def __setitem__(self, n, v):
        nns = self._pex(n)
        for nn in nns:
            self.pars[nn] = v

    def getPRef(self, n):
        return self.pars.getPRef(n)

    def freeParamVals(self):
        return self.pars.freeParamVals()

    def freeParamNames(self):
        return self.pars.freeParamNames()

    def freeParameters(self):
        return self.pars.freeParameters()

    def setFreeParamVals(self, vals):
        self.pars.setFreeParamVals(vals)

    def assignValues(self, vals):
        r"""
        Assign values to parameters

        Parameters
        ----------
        vals : dictionary
            Maps parameter name to new value
        """
        for k, v in six.iteritems(vals):
            self.pars[k] = v

    def _pex(self, pns):
        r"""
        Apply unix filename-like pattern matching and return updated list of parameters

        Returns
        -------
        updated list : list of strings
            Possibly expanded list of parameter names. For parameter names without expansion (and
            possibly without match in the list of model parameter names), the input string is returned.
        """
        # Ensure it is a list of strings
        if isinstance(pns, six.string_types):
            pns = [pns]
        r = []
        for p in pns:
            e = fnmatch.filter(list(self.parameters()), p)
            r.extend(e if len(e) > 0 else [p])
        return r

    def thaw(self, pns):
        pns = self._pex(pns)
        self.pars.thaw(pns)

    def freeze(self, pns):
        pns = self._pex(pns)
        self.pars.freeze(pns)

    def setRestriction(self, restricts, ap=True, usesmooth=False, sscale=1e-9):
        r"""
        Restrict parameter ranges

        Sets lower and upper bounds for parameters where desired. The bounds
        are saved to be used in constrained optimization, e.g., via
        fmin_cobyla. Additionally, a uniform prior is added to take effect
        on the posterior probability distribution. If desired, the edges
        of the priors are 'smoothed' to alleviate convergence problems with
        unconstrained optimization algorithms working on the posterior.

        Parameters
        ----------
        restricts : dictionary
            Maps parameter name to resctriction, i.e., a two-tuple giving
            lower and upper bound. Use None where no bound is desired.
        ap : boolean, optional
            Apply priors? If True (default), priors will be defined according
            to the restrictions.
        usesmooth : boolean, optional
            If True, a uniform prior with 'smoothed' edges is used instead
            if a sharply-edged uniform prior. The scale if the edge is given
            by `sscale`. Default is False.
        sscale : float, optional
            Scale of the edge used for smooth uniform prior (if desired).
        """
        self.pars.setRestriction(restricts)
        if ap:
            # Add prior(s)
            for k, v in six.iteritems(restricts):
                self.pars._checkParam(k)
                if any([not b is None for b in v]):
                    # Discard cases were both limits are none
                    if not usesmooth:
                        self.addUniformPrior(k, lower=v[0], upper=v[1])
                    else:
                        self.addSmoothUniformPrior(
                            k, lower=v[0], upper=v[1], scale=sscale)

    def getPenaltyFactor(self):
        return self._pf

    def setPenaltyFactor(self, pf):
        if pf < 0:
            raise (PE.PyAValError("Penalty factor must not be negative."))
        self._pf = pf

    def getRestrictions(self):
        return self.pars.getRestrictions()

    def parameters(self):
        return self.pars.parameters()

    def availableParameters(self):
        r"""
        Provides a list of existing parameter names.

        Returns
        -------
        Parameters : list of strings 
            A list with the names of available parameters.
        """
        return list(self.parameters())

    def frozenParameters(self):
        r"""
        Get names and values of frozen parameters.

        Returns
        -------
        Frozen parameters: dict
             Dictionary containing the names and values
             of all frozen parameters ({"parName":value, ...}).
        """
        fps = {}
        fpns = self.freeParamNames()
        for k, v in six.iteritems(self.parameters()):
            if not k in fpns:
                fps[k] = v
        return fps

    def setlogL(self, logl):
        """ Assign logL method """

        if isinstance(logl, six.string_types):
            # Specification by string
            if logl.lower() == "1dgauss":
                self.logL = types.MethodType(_gaussLogL, self)
            else:
                raise (PE.PyAValError("Unknown logl specification: " + str(logl)))
        elif hasattr(logl, "__call__"):
            # It is a callable
            self.logL = types.MethodType(logl, self)
        elif logl is None:
            # Do nothing here
            pass
        else:
            raise (PE.PyAValError("logl is neither a string nor a callable nor None.",
                                  where="MBO2"))

    def logL(self, *args, **kwargs):
        raise (PE.PyANotImplemented("""
hod 'logL' needs to be implemented.

logarithm of the likelihood for the
gatory arguments. Usually, however,
it requires a data set as input.
        """))

    def evaluate(self, *args, **kwargs):
        pass

    def logPrior(self, *args, **kwargs):
        """ Returns natural logarithm of prior for current parameter values """
        result = 0.0
        for p in self.priors:
            result += p.evaluate()
            if np.isinf(result):
                return result
        return result

    def _too(self, x, d):
        """ Returns d if x is None and x otherwise """
        if x is None:
            return d
        return x

    def addUniformPrior(self, pn, lower=None, upper=None):
        r"""
        Add a uniform prior

        Parameters
        ----------
        pn : string
            Name of the parameter
        lower, upper : float, optional
            Lower and upper bounds
        """
        self.pars._checkParam(pn)
        descr = "Uniform prior on '" + str(pn) + "' (lower = %g, upper = %g)" % (
            self._too(lower, -np.inf), self._too(upper, np.inf))
        self.addPrior(PyAUniformPrior(self.getPRef(
            pn), lower=lower, upper=upper, descr=descr))

    def addSmoothUniformPrior(self, pn, lower=None, upper=None, scale=1e-9):
        r"""
        Add a uniform prior with 'smoothed' (arctan) edges

        Parameters
        ----------
        pn : string
            Name of the parameter
        lower, upper : float, optional
            Lower and upper bounds
        """
        self.pars._checkParam(pn)
        descr = "Smooth uniform prior on '" + str(pn) + "' (lower = %g, upper = %g, scale = %g)" % \
                (self._too(lower, -np.inf), self._too(upper, np.inf), scale)
        self.addPrior(PyASmoothUniformPrior(self.getPRef(pn),
                      lower=lower, upper=upper, descr=descr))

    def addScalePrior(self, pn, m):
        r"""
        Add scale prior of the form x**m

        Parameters
        ----------
        pn : string
            Name of the parameter
        m : float
            Exponent (e.g., -1 for x**-1 prior distribution)
        """
        self.pars._checkParam(pn)
        descr = "Scale prior on '" + str(pn) + "' (exponent = %g)" % m
        self.addPrior(PyAScalePrior(self.getPRef(pn), m, descr=descr))

    def addCauchyPrior(self, pn, x0, scale, side):
        r"""
        Add Cauchy prior of the form scale / (pi*(scale**2 + (x-x0)**2)).

        The density is multiplied by two of a one-sided version is desired.

        Parameters
        ----------
        pn : string
            Name of the parameter
        x0 : float
            Location parameter of the distribution
        scale : float
            Scale parameter of the distribution
        side : string, {both, positive, negative}
            Use 'positive' to rule out values lower than x0 and negative for
            the reverse. Use 'both' to allow both sides of the center.
        """
        self.pars._checkParam(pn)
        descr = "Cauchy prior on '" + \
            str(pn) + "' (scale = %g, location = %g, side = %s)" % (scale, x0, side)
        self.addPrior(PyACauchyPrior(self.getPRef(pn),
                      scale, x0, side, descr=descr))

    def addNormalPrior(self, pn, mean, std):
        r"""
        Add normal (Gaussian) prior

        Parameters
        ----------
        pn : string
            Name of the parameter
        mean : float
            Location parameter of the distribution
        std : float
            Standard deviation of the distribution
        """
        self.pars._checkParam(pn)
        descr = "Normal prior on '" + \
            str(pn) + "' (mean = %g, std = %g)" % (mean, std)
        self.addPrior(PyANormalPrior(self.getPRef(pn), mean, std, descr=descr))

    def addPrior(self, pyaprior):
        r"""
        Add a prior

        Parameters
        ----------
        pyaprior : PyAPrior
            Instance of PyAPrior
        """
        self.priors.append(pyaprior)

    def mlD(self, *args, **kwargs):
        r"""
        Get marginal likelihood of the data (if known)

        Returns
        -------
        mlD : float
            Natural logarithm of the likelihood of the data if known
            and None otherwise.
        """
        return None

    def logPost(self, *args, **kwargs):
        r"""
        Get natural logarithm of the posterior

        Parameters
        ----------
        alllogs : boolean, optional
            If given and set True, also a tuple is returned, holding the logarithm
            of the priors and the likelihood separately.

        Returns
        -------
        lnp : float
            The (natural) logarithm of the posterior. The posterior is normalized if
            the marginal likelihood of the data is known and specified by the `mlD`
            method.
        pandl : tuple of floats, optional
            Tuple holding the natural logarithms of the priors and the likelihood.
            Only returned if alllogs is given as a keyword argument and True.
        """

        lp = self.logPrior(*args, **kwargs)
        if np.isinf(lp):
            if ("alllogs" in kwargs) and (kwargs["alllogs"] == True):
                return -np.inf, (-np.inf, -np.inf)
            else:
                return -np.inf

        ll = self.logL(*args, **kwargs)
        md = self.mlD(*args, **kwargs)
        po = lp + ll - (md or 0.0)

        if ("alllogs" in kwargs) and (kwargs["alllogs"] == True):
            return po, (lp, ll)

        return po

    def emceeLogPost(self, *args, **kwargs):
        r"""
        Get the logarithm of the posterior assuming that the first parameter is an array of free parameter values

        Sets the values of the free parameters to the specified values
        and evaluates the posterior probability. An implementation which
        can easily combined with sampling by emcee.

        Parameters
        ----------
        P : array
            An array holding values for the free parameters

        Returns
        -------
        lp : float
            Natural logarithm of the posterior
        """
        self.setFreeParamVals(args[0])
        kwargs["alllogs"] = True
        r = self.logPost(*args[1:], **kwargs)
        return r

    def addSPLikeObjf(self, f, name):
        r"""
        Add a SciPy-Like (SPL) objective function to the instance.

        Parameters
        ----------
        f : callable or string
            The scipy-like objective function to be assigned. A scipy-like objective function is one
            which takes as its first argument an array holding the values of the free parameters. The
            function may take any number of additional arguments and keywords.
        """

        if isinstance(f, six.string_types):
            if f == "-logl":
                def nln(self, *args, **kwargs):
                    """ -natural_log(Likelihood) """
                    return -self.logL(*args[1:], **kwargs)
                f = nln
            elif f == "-logpost":
                def nln(self, *args, **kwargs):
                    """ -natural_log(Posterior) """
                    return -self.logPost(*args[1:], **kwargs)
                f = nln
            elif f == "chisqr":
                f = getchisqrobjf()
            elif f == "nanchisqr":
                f = getchisqrobjf(sumfct=np.nansum)
            else:
                raise (PE.PyAValError("Unknown objective function string: '"+f+"'"))

        def objf(self, *args, **kwargs):
            # Make sure the current parameters are assigned
            self.setFreeParamVals(args[0])
            v = f(self, *args, **kwargs)
            if not np.isfinite(v):
                PE.PyAValError("Infinite value encountered in objective function for parameters: " +
                               str(args[0]) + ", free parameters : " + ",".join(self.freeParamNames()))
            v += self.getRPenalty()
            return v

        objf.__doc__ = f.__doc__
        setattr(self, name, types.MethodType(objf, self))

    def grad(self, *args, **kwargs):
        raise (PE.PyANotImplemented(
            "To use derivatives, implement the 'grad' function."))

    def getRPenalty(self):
        r"""
        Get penalty for violating restrictions

        Parameters
        ----------
        pf : float
            If a boundary, b, specified in a restriction is violated, the
            penalty is calculated as abs(b-v) x pf, where v is the current
            parameter value.

        Returns
        -------
        Penalty : float
            Penalty for current restrictions and parameter values
        """
        if self._pf == 0:
            return 0.0
        x = 0.0
        for p, r in six.iteritems(self.getRestrictions()):
            if (not r[0] is None) and (self[p] < r[0]):
                x += self._pf*abs(self[p]-r[0])
                continue
            if (not r[1] is None) and (self[p] > r[1]):
                x += self._pf*abs(self[p]-r[1])
                continue
        return x

#     def objfPenalize(self, pf=1e20):
#         """
#         Add restriction penalties to objective function
#         """
#         self._nonpenobjf = self.getSPLikeObjf()
#         def pobj(self, *args, **kwargs):
#             x = self._nonpenobjf(*args, **kwargs)
#             x += self.getRPenalty()
#             return x
#         pobj.__doc__ = self._nonpenobjf.__doc__ + " (penalized)"
#         self.objf = pobj


class MBOEv(MBO):
    r"""
    Model Base Object with an evaluate method and a predefined chi square objective function
    """

    def __init__(self, pars=None, rootName="", **kwargs):

        if pars is None:
            raise (PE.PyAValError("You need to specify the parameter names via 'pars'.",
                                  where="MBOEv",
                                  solution="Specify something along the lines of 'pars = ['pn1', 'pn2', ...]'."))

        MBO.__init__(self, pars, rootName, **kwargs)

        # Use likelihood based on Gaussian errors with std yerr
        self.setlogL("1dgauss")
        self.addSPLikeObjf("chisqr", "chisqr")
        self.addSPLikeObjf("nanchisqr", "nanchisqr")

    def evaluate(self, *args, **kwargs):
        r"""
        Evaluate model

        Parameters
        ----------

        """
        pass

    def _combineMBOs(self, right):
        r"""
        Combine two MBOs into a new one

        Parameters
        ----------
        right : MBO2 instance
            MBO object on the right hand side of the operation

        Returns
        -------
        combination : MBO2
        """
        r = MBOEv([], rootName="combined")
        r.pars = PyABaSoS.combine(self.pars, right.pars)
        r.leftCompo = self
        r.rightCompo = right
        return r


def _introdefarg(f, **kwargs):
    r"""
    Analyze and update keyword arguments of callable.

    Parameters
    ----------
    f : callable
        A function to be inspected (e.g., sco.fmin)
    kwargs : dictionary
        Dictionary of arguments to be updated.

    Returns
    -------
    kwargs, original kwargs : dictionaries
        Updated kwargs (including specifications from kwargs) and original
        kwargs as provided by the callable.
    """
    # Get dictionary of default arguments and default values
    fi = inspect.signature(f)
    odefargs = {param.name: param.default for param in fi.parameters.values(
    ) if param.default is not inspect.Parameter.empty}

    defargs = odefargs.copy()
    # Update entries for which kwargs holds values
    for k in list(defargs):
        if k in kwargs:
            defargs[k] = kwargs[k]
    return defargs, odefargs


def fitfmin(m, objf, x, y, yerr=None, **kwargs):
    r"""
    Use scipy's fmin to fit model.
    """
    # Get keywords and default arguments
    defargs, _ = _introdefarg(sco.fmin, **kwargs)

    if not yerr is None:
        defargs["args"] = (x, y, yerr)
    else:
        defargs["args"] = (x, y)

    defargs["full_output"] = True

    fr = sco.fmin(objf, m.freeParamVals(), **defargs)
    m.setFreeParamVals(fr[0])
    return fr


def fitfmin_cobyla(m, objf, x, y, yerr=None, cons=None, **kwargs):
    r"""
    Use scipy's fitfmin_cobyla to fit model.
    """
    # Get keywords and default arguments
    defargs, _ = _introdefarg(sco.fmin_cobyla, **kwargs)

    if not yerr is None:
        defargs["args"] = (x, y, yerr)
    else:
        defargs["args"] = (x, y)

    if cons is None:
        cons = []

    rs = m.getRestrictions()
    # Loop over freeParamNames (get order right)
    for i, p in enumerate(m.freeParamNames()):
        if p in rs:
            # There is a restriction for this parameter
            lower, upper = rs[p]
            # A function factory to close over 'j' (i.e., 'i' in the loop)
            # Constraints have to be positive if valid

            def getf(j):
                if (not lower is None) and (not upper is None):
                    # Upper and lower limit
                    def f(x): return min((x[j]-lower), (upper-x[j]))
                elif (not lower is None):
                    # Only lower limit
                    def f(x): return x[j]-lower
                elif (not upper is None):
                    # Only upper limit
                    def f(x): return upper-x[j]
                    # f = lambda *x:
                return f
            cons.append(getf(i))

    defargs["cons"] = cons
    # Otherwise, the constraint function will be called with the same
    # 'args' as the objective function
    defargs["consargs"] = ()

    fr = sco.fmin_cobyla(objf, m.freeParamVals(), **defargs)
    m.setFreeParamVals(fr)
    return fr


def fitfmin_powell(m, objf, x, y, yerr=None, **kwargs):
    r"""
    Use scipy's fmin_powell to fit model.
    """
    # Get keywords and default arguments
    defargs, _ = _introdefarg(sco.fmin_powell, **kwargs)

    if not yerr is None:
        defargs["args"] = (x, y, yerr)
    else:
        defargs["args"] = (x, y)

    defargs["full_output"] = True

    fr = sco.fmin_powell(objf, m.freeParamVals(), **defargs)
    m.setFreeParamVals(fr[0])
    return fr


def fitfmin_l_bfgs_b(m, objf, x, y, yerr=None, **kwargs):
    r"""
    Use scipy's fmin_l_bfgs_b to fit model.

    Parameters
    ----------
    userBounds : dictionary, optional
        Maps parameter name to a tuple of the form (lower, upper) bound.
        If given, the bounds given there will overrule those specified
        in the form of restrictions (should there by any).
    """
    # Get keywords and default arguments
    defargs, _ = _introdefarg(sco.fmin_l_bfgs_b, **kwargs)

    if not yerr is None:
        defargs["args"] = (x, y, yerr)
    else:
        defargs["args"] = (x, y)

    userBounds = kwargs.get("userBounds") or {}

    rs = m.getRestrictions()
    bounds = []
    # Loop over freeParamNames (get order right)
    for i, p in enumerate(m.freeParamNames()):
        if p in userBounds:
            bounds.append(userBounds[p])
        elif p in rs:
            # There is a restriction for this parameter
            bounds.append(rs[p])
        else:
            bounds.append((None, None))

    defargs["bounds"] = bounds
    defargs["approx_grad"] = True

    fr = sco.fmin_l_bfgs_b(objf, m.freeParamVals(), **defargs)
    m.setFreeParamVals(fr[0])
    return fr


def sampleEMCEE2(m, pargs=(), walkerdimfac=4, scales=None,
                 sampleArgs=None, dbfile="chain.emcee", ps=None, emcp=None, toMAP=True):
    r"""
    MCMC sampling using emcee package.

    Sample from the posterior probability distribution using the emcee
    package. 

    The emcee sampler can be accessed via the `emceeSampler` attribute,
    which may be used to continue or manipulate sampling.

    Parameters
    ----------
    walkerdimfac : int, optional
        Determines the number of walkers. By default, four times the
        number of free parameters is used.
    pargs : tuple
        Content of 'args' argument handed to EnsembleSampler
    scales : dictionary, optional
        The scales argument can be used to control the initial distribution
        of the walkers. By default, all walkers are distributed around the
        location given by the current state of the object, i.e., the current
        parameter values. In each direction, the walker are randomly distributed
        with a Gaussian distribution, whose default standard deviation is 1e-8.
        The scales argument can be used to control the width of Gaussians used
        to distribute the walkers.
    sampleArgs : dictionary, optional
        Keyword controlling the sampling process. Use 'burn' (int) to specify
        the number of burn-in iterations (default is 0). Via 'iters' (int)
        the numbers of iterations after the burn-in can be specified (default 1000).
        The 'process' (int) key can be used to control the number of iterations after
        which the progress bar is updated (default is iters/100). Note that the
        'progressbar' package must be installed to get a progress bar. Otherwise
        more mundane print statements will be used.  
    dbfile : string, optional
        The result of the sampling, i.e., the chain(s), the corresponding
        values of the posterior, and the names of the free parameters are
        saved to the specified file (by default 'chain.emcee' is used).
        The traces stored there can be analyzed using the 'TraceAnalysis'
        class. Set this parameter to 'None' to avoid saving the results.
    ps : tuple, optional
        A tuple holding the current position and state of the sampler. This
        tuple is returned by this method. The `ps` argument can be used
        to continue sampling from the last state. Note that no burn-in will
        be carried out and the other arguments should be given as previously
        to continue sampling successfully. 
    emcp : dictionary, optional
        Extra arguments handed to `EnsembleSampler` object.
    toMAP : boolean, optional
        If True (default), the object is set to the maximum posterior probability point sampled.
        Otherwise, it remains in a random state.

    Returns
    -------
    (pos, state) : tuple
        Position and state of the sampler. Can be used to continue sampling.
    sampler : emceeSampler
        The emcee object used for sampling.
    """

    if not ic.check["emcee"]:
        raise (PE.PyARequiredImport("Could not import the 'emcee' package.",
                                    solution="Please install 'emcee'."))

    # Names of free parameters
    fpns = m.freeParamNames()
    # Values of of free parameters
    fpvs = m.freeParamVals()
    # Names and values of free parameter (in no order)
    fps = dict(zip(fpns, fpvs))
    # Number of dimensions
    ndims = len(fpns)

    if ndims == 0:
        raise (PE.PyAValError("At least one free parameter is required for sampling.",
                              where="sampleEMCEE2",
                              solution="Use 'thaw' to free same parameters."))

    if not dbfile is None:
        if re.match(r".*\.emcee$", dbfile) is None:
            PE.warn(PE.PyAValError("The db filename (" + str(dbfile) + ") does not end in .emcee. TraceAnalysis will not recognize it as an emcee trace file.",
                                   solution="Use a filename of the form *.emcee"))

    # Number of walkers
    nwalker = ndims * walkerdimfac

    if nwalker < ndims * 2:
        raise (PE.PyAValError("The number of walkers must be at least twice the number of free parameters.",
                              where="sampleEMCEE2",
                              solution="Increase the number of walkers."))
    if nwalker % 2 == 1:
        raise (PE.PyAValError("The number of walkers must be even.",
                              where="sampleEMCEE2",
                              solution="Use an even number of walkers."))

    # Set default values for sampleArgs
    if sampleArgs is None:
        sampleArgs = {}
    if not "burn" in sampleArgs:
        sampleArgs["burn"] = 0
    if not "iters" in sampleArgs:
        sampleArgs["iters"] = 1000
    if not "progress" in sampleArgs:
        sampleArgs["progress"] = sampleArgs["iters"] / 100

    if ps is None:

        if emcp is None:
            emcp = {}

        # Generate the sampler
        emceeSampler = emcee.EnsembleSampler(
            nwalker, ndims, m.emceeLogPost, args=pargs, **emcp)

        if scales is None:
            scales = {}

        # Generate starting values
        pos = []
        for _ in smo.range(nwalker):
            pos.append(np.zeros(ndims))
            for i, n in enumerate(fpns):
                # Use 1e-8 unless scale is defined
                s = (scales.get(n) or 1e-8)
                pos[-1][i] = np.random.normal(fps[n], s)

        # Default value for state
        state = None

        if sampleArgs["burn"] > 0:
            # Run burn-in
            pos, prob, state, blobs = emceeSampler.run_mcmc(
                pos, sampleArgs["burn"])
            # Reset the chain to remove the burn-in samples.
            emceeSampler.reset()

    else:
        # Assign position and state from previous run
        pos, state = ps

    if (not sampleArgs["progress"] is None) and ic.check["progressbar"]:
        widgets = ['EMCEE progress: ', progressbar.Percentage(), ' ', progressbar.Bar(marker=progressbar.RotatingMarker()),
                   ' ', progressbar.ETA()]
        pbar = progressbar.ProgressBar(
            widgets=widgets, maxval=sampleArgs["iters"]).start()

    n = 0
    for pos, prob, state, blobs in emceeSampler.sample(pos, rstate0=state, iterations=sampleArgs["iters"], thin=1):
        n += 1
        if (not sampleArgs["progress"] is None) and (n % sampleArgs["progress"] == 0):
            if ic.check["progressbar"]:
                pbar.update(n)
            else:
                print("EMCEE: Reached iteration ",
                      n, " of ", sampleArgs["iters"])

    # Save the chain to a file
    if not dbfile is None:

        def deblob(b):
            """ Convert blob format (list of lists) into numpy arrays """
            # Shape is (no. of walkers, length of chain)
            lnprior = np.zeros((len(b[0]), len(b)))
            lnl = np.zeros((len(b[0]), len(b)))
            for i, blob in enumerate(b):
                lnprior[::, i] = np.array([x[0] for x in blob])
                lnl[::, i] = np.array([x[1] for x in blob])
            return lnprior, lnl

        lnprior, lnl = deblob(emceeSampler.blobs)
        np.savez_compressed(open(dbfile, 'wb'), chain=emceeSampler.chain, lnpost=emceeSampler.lnprobability, lnprior=lnprior, lnl=lnl,
                            pnames=np.array(fpns, dtype=np.str_))

    if toMAP:
        # Set to Maximum-A-Posteriori solution
        indimin = np.argmax(emceeSampler.lnprobability)
        for i, p in enumerate(fpns):
            m[p] = emceeSampler.flatchain[indimin, i]

    return (pos, state), emceeSampler


def _fuf2ModelFactory(odc, rn):

    class F2M(MBOEv):

        def __init__(self, *args, **kwargs):
            self._odm = odc(*args, **kwargs)
            MBOEv.__init__(self, pars=list(
                self._odm.parameters()), rootName=rn)
            self.setlogL("1dgauss")

        def evaluate(self, *args, **kwargs):
            self._odm.assignValues(self.parameters())
            return self._odm.evaluate(args[0])

    return F2M


# Transform OneDFit objects into MBOs
# List gives tuples with fuf2-name, fuf-class, and root name to be assigned
_ml = [("GaussFit1d", _fuf.GaussFit1d, "GF"),
       ("MultiGaussFit1d", _fuf.MultiGauss1d, "MGF"),
       ("CauchyLorentz1d", _fuf.CauchyLorentz1d, "CauchyLorentz"),
       ("Voigt1d", _fuf.voigt1d, "Voigt"),
       ("MultiVoigt1d", _fuf.MultiVoigt1d, "MultiVoigt"),
       ("SinusFit1d", _fuf.SinusFit1d, "SinusFit"),
       ("ExpDecayFit1d", _fuf.ExpDecayFit1d, "ExpDecayFit"),
       ("PolyFit1d", _fuf.PolyFit1d, "PolyFit"),
       ("ConstantFit1d", _fuf.ConstantFit1d, "ConstantFit"),
       ("GaussFit2d", _fuf.GaussFit2d, "GaussFit2d"),
       ("MultiGauss2d", _fuf.MultiGauss2d, "MultiGauss2d"),
       ("Circle2d", _fuf.Circle2d, "Circle2d")]

# Provide versions/names without the 1d ending
_eml = []
for m in _ml:
    if m[0].endswith("1d"):
        _eml.append((m[0][0:-2], m[1], m[2]))
_ml.extend(_eml)

# Add models from modelSuite
_ml.extend([("PalLC", _ms.palTrans.PalLC, "PalLC"),
            ("PalLCKep", _ms.palTrans.PalLCKep, "PalLCKep"),
            ("MandelAgolLC", _ms.forTrans.MandelAgolLC, "MandelAgolLC"),
            ("MandelAgolNLLC", _ms.forTrans.MandelAgolNLLC, "MandelAgolNLLC"),
            ("LimBrightTrans", _ms.LimBrightTrans, "LimBrightTrans"),
            ("RmcL", _ms.RmcL, "RMcL"),
            ("RmcLell", _ms.RmcLell, "RmcLell"),
            ("SinRadVel", _ms.SinRadVel, "SinRadVel"),
            ("KeplerEllipseModel", _ms.keplerEllipseModel, "KeplerEllipseModel"),
            ("KeplerRVModel", _ms.KeplerRVModel, "KeplerRVModel"),
            ("LLGauss", _ms.LLGauss, "LLGauss"),
            ("VoigtAstroP", _ms.VoigtAstroP, "VoigtAstroP"),
            ("LyaTransmission", _ms.LyaTransmission, "LyaTransmission"),
            ("RotBroadProfile", _ms.RotBroadProfile, "RotBroadProfile")
            ])

for m in _ml:
    setattr(_fbo2module, m[0], _fuf2ModelFactory(m[1], m[2]))


class Poly2d(MBOEv):
    r"""
    Bivariate (2d) polynomial model

    Parameters
    ----------
    nx, ny : int
        Number of coefficients in first and second variable
    """

    def __init__(self, nx, ny):
        if (nx < 0) or (ny < 0):
            raise (PE.PyAValError("nx and ny must be non-negative integers."))
        self.nx, self.ny = nx, ny
        pars = []
        for x in range(nx+1):
            for y in range(ny+1):
                pars.append("c%d-%d" % (x, y))
        # 'pars' specifies parameter names in the model
        MBOEv.__init__(self, pars=pars, rootName="Poly2d")

    def evaluate(self, x, *args):
        r"""
        Evaluate model at points (x[::,0], x[::,1])
        """
        c = np.zeros((self.nx+1, self.ny+1))
        for i in range(self.nx+1):
            for j in range(self.ny+1):
                c[i, j] = self["c%d-%d" % (i, j)]
        return np.polynomial.polynomial.polyval2d(x[::, 0], x[::, 1], c)

#
# # class GF1d(MBO2):
# #
# #     def __init__(self):
# #         self._gf = GaussFit1d()
# #         MBO2.__init__(self, pars=list(self._gf.parameters()), rootName=GF1d)
# #         self.setLogL("1dgauss")
# #
# #     def evaluate(self, t):
# #         self._gf.assignValues(self.parameters())
# #         return self._gf.evaluate(t)
# #
# #
# #
# class Poly2(MBO2):
#
#     def __init__(self):
#         MBO2.__init__(self, pars=["c0", "c1", "c2"], rootName="Poly2")
#         self.setlogL("1dgauss")
#
#     def evaluate(self, x, **kwargs):
#         s = self._imap
#         return s["c0"] + s["c1"]*x + s["c2"]*x**2
#
#
# class CeleriteModel(MBO2):
#
#     def _nc(self, s):
#         """
#         Adjust parameter names
#
#         Parameters
#         ----------
#         s : string
#             Parameter name from celerite
#
#         Returns
#         -------
#         name : string
#             Adjusted name (here, 'kernel: ' removed)
#         """
#         s = s.replace("kernel:", "")
#         return s
#
#     def __init__(self, gp, nc=None):
#
#         # Name conversion
#         if nc is None:
#             # Use default naming convention
#             nc = self._nc
#
#         # Save reference to GP
#         self.gp = gp
#         # Save parameter names (and the order)
#         self._pns = [nc(n) for n in gp.get_parameter_names(include_frozen=True)]
#
#         MBO2.__init__(self, self._pns, rootName="celmo", logl=None)
#
#         vals = gp.get_parameter_vector(include_frozen=True)
#         for i, n in enumerate(self._pns):
#             self[n] = vals[i]
#
#         for i, b in enumerate(self.gp.get_parameter_bounds(include_frozen=True)):
#             self.setRestriction({self._pns[i]:b})
#
#     def _updateGPParams(self):
#         """ Assign current model parameters """
#         pv = np.zeros(len(self._pns))
#         for i, _ in enumerate(self.gp.get_parameter_names()):
#             pv[i] = self[self._pns[i]]
#         self.gp.set_parameter_vector(pv, include_frozen=True)
#
#     def logL(self, *args, **kwargs):
#         """ Likelihood function for celerite GP """
#         self._updateGPParams()
#         ll = self.gp.log_likelihood(args[1])
#         return ll
#
#     def evaluate(self, x):
#         return None
#
#     def predict(self, px, y):
#         """
#         Get predicted mean and variance of GP
#
#         Parameters
#         ----------
#         px : array
#             Position at which to evaluate the prediction
#         y : array
#             The data
#
#         Returns
#         -------
#         mean, variance, std : arrays
#             Prediction
#         """
#         self._updateGPParams()
#         pred_mean, pred_var = self.gp.predict(y, px, return_var=True)
#         return pred_mean, pred_var, np.sqrt(pred_var)
#
#



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import _ic
import numpy as np


def airtovac(wave, depWarn=True):
    """
    Convert air wavelengths to vacuum wavelengths

    .. warning::
       The conversion implemented here is based on the older formulae
       given by Edlen 1953. Furthermore, it seems that wave numbers in air
       are used, where vacuum wave numbers should be used, which, however,
       produces only a second-order deviation. Consider using
       :py:func:`airtovac2` instead.

    Parameters
    ----------
    wave : float, array
        The wavelength in air [Angstrom]
    depWarn : boolean, optional
        If True (default), a deprecation warning will be
        given.

    Returns
    -------
    Wavelength : array
        Wavelength in vacuum [Angstrom]

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          AIRTOVAC
    PURPOSE:
          Convert air wavelengths to vacuum wavelengths
    EXPLANATION:
          Wavelengths are corrected for the index of refraction of air under
          standard conditions.  Wavelength values below 2000 A will not be
          altered.  Uses the IAU standard for conversion given in Morton
          (1991 Ap.J. Suppl. 77, 119)

    CALLING SEQUENCE:
          AIRTOVAC, WAVE

    INPUT/OUTPUT:
          WAVE - Wavelength in Angstroms, scalar or vector
                 WAVE should be input as air wavelength(s), it will be
                 returned as vacuum wavelength(s).  WAVE is always converted to
                 double precision upon return.

    EXAMPLE:
          If the air wavelength is  W = 6056.125 (a Krypton line), then
          AIRTOVAC, W yields an vacuum wavelength of W = 6057.8019

    METHOD:
          See Morton (Ap. J. Suppl. 77, 119) for the formula used

    REVISION HISTORY
          Written W. Landsman                   November 1991
          Converted to IDL V5.0   W. Landsman   September 1997
    """

    if depWarn:
        PE.warn(
            PE.PyADeprecationError(
                "Note: airtovac is outdated; see documentation for reasons.",
                solution="Consider using 'airtovac2'.",
            )
        )

    sigma2 = (1.0e4 / wave) ** 2.0  # Convert to wavenumber squared

    # Compute conversion factor

    fact = 1.0 + 6.4328e-5 + 2.94981e-2 / (146.0 - sigma2) + 2.5540e-4 / (41.0 - sigma2)

    fact = fact * (wave >= 2000.0) + 1.0 * (wave < 2000.0)

    wave = wave * fact  # Convert Wavelength

    return wave


def vactoair(wave, depWarn=True):
    """
    Convert vacuum wavelengths to air wavelengths

    .. warning::
       The conversion implemented here is based on the older formulae
       given by Edlen 1953. Furthermore, it seems that wave numbers in air
       are used, where vacuum wave numbers should be used, which, however,
       produces only a second-order deviation. Consider using
       :py:func:`vactoair2` instead.

    Parameters
    ----------
    wave : float, array
        The wavelength in vacuum [Angstrom]

    Returns
    -------
    Wavelength : array,
        Wavelength in air [Angstrom]
    depWarn : boolean, optional
        If True (default), a deprecation warning will be
        given.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          VACTOAIR
    PURPOSE:
          Convert vacuum wavelengths to air wavelengths
    EXPLANATION:
          Corrects for the index of refraction of air under standard conditions.
          Wavelength values below 2000 A will not be altered.  Accurate to
          about 0.005 A

    CALLING SEQUENCE:
          VACTOAIR, WAVE

    INPUT/OUTPUT:
          WAVE - Wavelength in Angstroms, scalar or vector
                  WAVE should be input as vacuum wavelength(s), it will be
                  returned as air wavelength(s).  WAVE is always converted to
                  double precision

    EXAMPLE:
          If the vacuum wavelength is  W = 2000, then

          IDL> VACTOAIR, W

          yields an air wavelength of W = 1999.353 Angstroms

    METHOD:
          An approximation to the 4th power of inverse wavenumber is used
          See IUE Image Processing Manual   Page 6-15.

    REVISION HISTORY
          Written, D. Lindler 1982
          Documentation W. Landsman  Feb. 1989
          Converted to IDL V5.0   W. Landsman   September 1997
    """

    if depWarn:
        PE.warn(
            PE.PyADeprecationError(
                "Note: vactoair is outdated; see documentation for reasons.",
                solution="Consider using 'vactoair2'.",
            )
        )

    wave2 = wave**2.0
    fact = 1.0 + 2.735182e-4 + 131.4182 / wave2 + 2.76249e8 / (wave2**2.0)
    fact = fact * (wave >= 2000.0) + 1.0 * (wave < 2000.0)

    # Convert wavelengths

    wave = wave / fact

    return wave


def specAirVacConvert(wvl, flux, direction, dontUse2=False):
    """
    Transform spectrum between air and vacuum wavelengths.

    In contrast to :py:func:`airtovac` and :py:func:`vactoair`,
    which only transform the wavelength axis, this function
    converts the wavelength axis and interpolates the shifted
    spectrum to retain the input wavelength axis.

    Due to the shift, a fraction of the spectrum is lost on one
    side, while there is no valid information on the other.
    Therefore, the function returns both a new flux array obtained
    by linear interpolation and an array of valid indices, i.e.,
    those indices in the new spectrum, which are not NaN.

    .. warning::
       Linear interpolation of the spectrum can affect its
       noise properties. This function
       does *nothing* to prevent or correct for that effect.

    Parameters
    ----------
    wvl : array
        Input wavelength axis.
    flux : array
        Input flux array.
    direction : string, {"airtovac","vactoair"}
        If the direction is "vactoair", the input spectrum is
        assumed to be obtained in vacuum conditions. The opposite
        is true otherwise.
    dontUse2 : boolean, optional
        If True, the now deprecated function airtovac and vactoair
        will be used. Otherwise (default), airtovac2 and vactoair2
        will be used with default settings.

    Returns
    -------
    New flux : array
        The shifted flux array, defined on the input wavelength axis.
    Valid indices : array
        The array of valid, i.e., not NaN, indices.
    """
    if not _ic.check["scipy"]:
        raise (
            PE.PyARequiredImport(
                "This function requires scipy.", solution="Install scipy."
            )
        )
    import scipy.interpolate as sci

    if direction == "airtovac":
        if dontUse2:
            newwvl = airtovac(wvl)
        else:
            newwvl = airtovac2(wvl)
    elif direction == "vactoair":
        if dontUse2:
            newwvl = vactoair(wvl)
        else:
            newwvl = vactoair2(wvl)
    else:
        raise (
            PE.PyAValError(
                "The direction '" + str(direction) + "' is not supported.",
                solution="Choose either 'airtovac' or 'vactoair'.",
            )
        )

    # Use spectrum with converted wavelength axis for interpolation
    shiftSpec = sci.interp1d(newwvl, flux, bounds_error=False, fill_value=np.nan)
    # Obtain flux on 'old' wvl axis by interpolation
    sflux = shiftSpec(wvl)
    # Determine 'valid' indices
    vind = np.where(np.bitwise_not(np.isnan(sflux)))[0]

    return sflux, vind


def _iocon(inp):
    """
    Convert input into np float array and create function to convert back.

    Parameters
    ----------
    inp : some numeric type
        Input

    Returns
    -------
    output : float array
        The input converted into a numpy float array
    recovery function : callable
        Function that converts the output back into a float
        number unless the input was an numpy array. In the latter
        case, it remains an array.
    """

    try:
        # Using ravel to avoid generating object with zero length
        out = np.ravel(np.array(inp, dtype=float))
    except Exception as e:
        raise (
            PE.PyAValError(
                "Could not convert input of type: "
                + str(type(inp))
                + " into float array. Error: "
                + str(e),
                where="_IOCON",
                solution="Use an input which can be converted into a numpy float array.",
            )
        )

    if isinstance(inp, np.ndarray) or len(out) > 1:
        return out, lambda x: x
    else:

        def recover(x):
            try:
                result = float(x)
            except Exception as e:
                raise (
                    PE.PyAValError(
                        "Could not convert result into a float. Error: " + str(e),
                        where="_iocon",
                        solution="Use a float or numpy array as input. Are you maybe using a list?",
                    )
                )
            return result

        return out, recover


class RefractiveIndexAV:
    """
    Refractive index for conversion between air- and vacuum wavelengths.

    The conversion is based on the formulae provided either by:
      - Edlen 1953, J. Opt. Soc. Am 43 no. 5
      - Peck and Reeder 1972, J. Opt. Soc. 62 no. 8
      - Ciddor 1996, Applied Optics 35 no. 9

    This class provides the routine `refractiveIndex`, `vactoair`,
    and `airtovac` to convert between wavelengths in air and vacuum.

    The method `refractiveIndex` points to either of `_edlen53`,
    `_ciddor`, or `_peckReeder` depending on the mode chosen.

    To conversion from vacuum to air wavelengths is given by the
    respective formulae provided by the authors of the above papers, who
    all specify the refractive index of "standard air" as a function
    of vacuum wave number. The reverse conversion from air into vacuum
    is done by applying an iterative scheme. Standard air is at 15 degrees
    Celsius, dry, and at a pressure of 750 mmHg = 101325 Pa. Ciddor assumes
    a CO2 concentration of 450 ppm, while Edlen, and Peck and Reeder have
    assumed 300 ppm.

    Parameters
    ----------
    mode : string, {edlen53, peckReeder, ciddor}, optional
        The source for the wavelength conversion. The default is
        "ciddor".

    """

    def _edlen53(self, wvl):
        """
        Refractive index according to Edlen 1953.

        Parameters
        ----------
        wvl : float or array
            Vacuum wavelength [A]

        Returns
        -------
        n : float or array
            The refractive index.
        """
        if np.any(wvl < 2000.0):
            raise (
                PE.PyAValError(
                    "The Edlen 1953 conversion covers only the >2000 A range.",
                    where="RefractiveIndexAV::_edlen53",
                    solution="Adapt the wavelength range",
                )
            )

        s2 = (1e4 / wvl) ** 2
        n = 1.0 + 6.4328e-5 + (2.94981e-2 / (146.0 - s2)) + (2.554e-4 / (41.0 - s2))
        return n

    def _ciddor(self, wvl, xc=None):
        """
        Refractive index given by Ciddor 1996.

        Parameters
        ----------
        wvl : float or array
            Vacuum wavelength [A]
        xc : float, optional
            Concentration of CO2 [ppm]

        Returns
        -------
        n : float or array
            Refractive index
        """
        if np.any(wvl < 3000.0) or np.any(wvl > 16900.0):
            raise (
                PE.PyAValError(
                    "The Ciddor conversion covers only the 3000 - 16900 A range.",
                    where="RefractiveIndexAV::_ciddor",
                    solution="Adapt the wavelength range",
                )
            )

        # Constants in [(10-6 m)**-2] from Ciddor 1996, Applied Optics, Vol. 35, Issue 9, pp. 1566-1573
        # Appendix A
        k0 = 238.0185
        k1 = 5792105.0
        k2 = 57.362
        k3 = 167917.0

        s2 = (1e4 / wvl) ** 2

        # Eqs. 1 and 2
        nas = (k1 / (k0 - s2) + k3 / (k2 - s2)) / 1e8 + 1.0
        if xc is not None:
            naxs = (nas - 1.0) * (1.0 + 0.534e-6 * (xc - 450.0)) + 1.0
            return naxs
        return nas

    def _peckReeder(self, wvl):
        """
        Refractive index from Peck and Reeder 1972.

        Parameters
        ----------
        wvl : float or array
            Vacuum wavelength [A]

        Returns
        -------
        n : float or array
            The refractive index.
        """
        if np.any(wvl < 2300.0) or np.any(wvl > 16900.0):
            raise (
                PE.PyAValError(
                    "The Peck and Reeder conversion covers only the 2300 - 16900 A range.",
                    where="RefractiveIndexAV::_peckReeder",
                    solution="Adapt the wavelength range",
                )
            )

        s2 = (1e4 / wvl) ** 2

        # Eq. 2
        n = (5791817.0 / (238.0185 - s2) + 167909.0 / (57.362 - s2)) / 1e8 + 1.0
        return n

    def vactoair(self, wvl, **kwargs):
        """
        Convert vacuum wavelength into wavelength in air.

        Parameters
        ----------
        wvl : float or array
            Wavelength in vacuum [A].
        kwargs : dict
            Additional keywords eventually needed in the call
            to the refractiveIndex function (depends on the
            specified 'mode').

        Returns
        -------
        wvl : float or array
            Wavelength in air.
        """
        wvl, rec = _iocon(wvl)
        result = wvl / self.refractiveIndex(wvl, **kwargs)
        return rec(result)

    def airtovac(self, wvl, precision=1e-12, maxiter=30, **kwargs):
        """
        Convert air wavelength into wavelength in vacuum.

        An iterative scheme is applied to invert the formula.

        Parameters
        ----------
        wvl : float or array
            Wavelength in vacuum [A].
        precision : float, optional
            The target precision beyond which iteration should
            be stopped; default is 1e-12.
        maxiter : int, optional
            Maximum number of iterations used to invert the
            conversion; default is 30.
        kwargs : dict
            Additional keywords eventually needed in the call
            to the refractiveIndex function (depends on the
            specified 'mode').

        Returns
        -------
        wvl : float or array
            Wavelength in air.
        """
        # First guess for vacuum wavelength is wavelength in air
        wvl, rec = _iocon(wvl)
        lvac = wvl.copy()
        counter = 0
        while True:
            counter += 1
            dl = lvac / self.refractiveIndex(lvac) - wvl
            lvac -= dl
            # Obtain difference between air wavelength resulting from
            # current estimate of vacuum wavelength
            dlair = self.vactoair(lvac) - wvl
            if np.abs(np.max(dlair)) < precision:
                return rec(lvac)
            if counter > maxiter:
                p = np.max(dlair)
                raise (
                    PE.PyAValError(
                        "Maximum of iterations reached. Current precision is "
                        + str(p)
                        + " A.",
                        where="RefractiveIndexAV::airtovac",
                        solution=[
                            "Try to increase 'maxiter'",
                            "Try to use subrange of wavelengths",
                            "Try to adapt 'precision' to necessary value",
                        ],
                    )
                )

        return rec(lvac)

    def __init__(self, mode="ciddor"):
        if mode == "edlen53":
            self.refractiveIndex = self._edlen53
        elif mode == "peckReeder":
            self.refractiveIndex = self._peckReeder
        elif mode == "ciddor":
            self.refractiveIndex = self._ciddor
        else:
            raise (
                PE.PyAValError(
                    "Unknown mode: '" + str(mode) + "'.",
                    where="RefractiveIndexAV",
                    solution="Use either of 'edlen53', 'peckReeder', or 'ciddor'.",
                )
            )


def vactoair2(wvl, mode="ciddor", **kwargs):
    """
    Converts wavelength in vacuum into wavelength in air.

    This function is a wrapper around the `RefractiveIndexAV` class.

    Parameters
    ----------
    wvl :float or array
        Wavelength in vacuum [A].
    mode : string, {edlen53, peckReeder, ciddor}, optional
        The source for the wavelength conversion. The default is
        "ciddor".
    kwargs : dict
        Additional parameters handed to the vactoair method
        of RefractiveIndexAV.

    Returns
    -------
    wvl : float or array
        Wavelength in air [A].
    """
    c = RefractiveIndexAV(mode=mode)
    return c.vactoair(wvl, **kwargs)


def airtovac2(wvl, mode="ciddor", **kwargs):
    """
    Converts wavelength in air into wavelength in vacuum.

    This function is a wrapper around the `RefractiveIndexAV` class.

    Parameters
    ----------
    wvl :float or array
        Wavelength in vacuum [A].
    mode : string, {edlen53, peckReeder, ciddor}, optional
        The source for the wavelength conversion. The default is
        "ciddor".
    kwargs : dict
        Additional parameters handed to the airtovac method
        of RefractiveIndexAV.

    Returns
    -------
    wvl : float or array
        Wavelength in vacuum [A].
    """
    c = RefractiveIndexAV(mode=mode)
    return c.airtovac(wvl, **kwargs)



from .voigtAstro import VoigtAstroP
from PyAstronomy import funcFit as fuf
from PyAstronomy.pyasl import convertDampingConstant
import numpy as np

class LyaTransmission(fuf.OneDFit):
  """
    Lyman alpha transmission profile including Deuterium absorption.
    
    The transmission is given by
      
    .. math::
       
        e^{-\\sigma N}\\, ,
        
    where N is the column density and :math:`\\sigma` is the wavelength-dependent
    cross-section.
    
    *Fit Parameters*:
    
    ===== ============================= =====
    N     Hydrogen column density       /cm^2
    b     Doppler parameter             km/s
    Dfrac Deuterium fraction            --
    ===== ============================= =====
    
    Parameters
    ----------
    N : float, optional
        Hydrogen column density [/cm^2]. The default is 0.
    b  : float, optional
        The Doppler parameter (corresponds to sqrt(2) times
        the velocity dispersion) to model thermal line width [km/s].
        The default is 10 km/s.
    D_fraction : float, optional
        Fractional abundance of Deuterium with respect to Hydrogen.
        The default is 1.5e-5.
  """
  
  def __init__(self, N=0.0, b=10.0, D_fraction=1.5e-5):
    fuf.OneDFit.__init__(self, ["N","b", "Dfrac"], rootName="LyaTransmission")
    self["b"] = b
    self["Dfrac"] = D_fraction
    # (Only) Einstein coefficient relevant for LyA
    elya = 6.258085e8
    # Hydrogen
    self._absH = VoigtAstroP()
    self._absH["w0"] = 1215.67
    self._absH["b"] = self["b"]
    self._absH["gamma"] = convertDampingConstant(elya, 1215.67)
    self._absH["f"] = 0.416
    # Deuterium, its width is sqrt(2) times smaller
    self._absD = VoigtAstroP()
    self._absD["w0"] = 1215.34
    self._absD["b"] = self["b"]/np.sqrt(2.0)
    self._absD["gamma"] = convertDampingConstant(elya, 1215.67)
    self._absD["f"] = 0.416
  
  def evaluate(self, x):   
    """
      Evaluate the transmission profile.
      
      Parameters
      ----------
      x : array of floats
          Contains the wavelength values in Angstrom.
      
      Returns
      -------
      Model : array of floats
          The line profile.
    """
    self._absH["b"] = self["b"]
    self._absD["b"] = self["b"]/np.sqrt(2.0)
    y = np.exp(-self["N"] * self._absH.evaluate(x)) * \
        np.exp(-self["N"] * self["Dfrac"] * self._absD.evaluate(x))
    return y



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
import copy
import numpy as np


class FufDS:
    r"""
    FuncFit data set.

    Parameters
    ----------
    x : array
        Specify the abscissa.
    y : array
        Specify the data.
    yerr : array, optional
        The error of the data.

    """

    def __init__(self, x, y, yerr=None):
        if not isinstance(y, np.ndarray):
            raise (PE.PyAValError("`y` must be a numpy.ndarray. Current type is '" + str(type(y)) + "'.",
                                  solution="Make `y` a numpy array."))
        self.y = y.copy()
        if yerr is not None:
            if not isinstance(yerr, np.ndarray):
                raise (PE.PyAValError("`yerr` must be a numpy.ndarray. Current type is '" + str(type(y)) + "'.",
                                      solution="Make `yerr` a numpy array."))
                if y.shape != yerr.shape:
                    raise (PE.PyAValError("`y` and `yerr` must have the same shape.",
                                          solution="Check and adapt shapes of arrays."))
            self.yerr = yerr.copy()
        else:
            self.yerr = None
        if isinstance(x, np.ndarray):
            self.x = x.copy()
        else:
            self.x = copy.copy(x)

    def yerrDefined(self):
        r"""
        Check whether yerr is defined or None.

        Returns
        -------
        Defined : boolean
            True, if yerr is not None.
        """
        return not (self.yerr is None)

    def xyyerrDefined(self):
        r"""
        Check whether x, y, and yerr are not None:

        Returns
        -------
        Defined : boolean
            True, if x, y, and yerr are not None. 
        """
        xnn = not (self.x is None)
        ynn = not (self.y is None)
        yenn = self.yerrDefined()
        return (xnn & ynn & yenn)



from PyAstronomy.funcFit import OneDFit
import numpy

class GeomPlanetBrightPhase(OneDFit):
  
  def __init__(self):
    """
      @TODO - docu
    """
    OneDFit.__init__(self, ["i", "per", "T0", "brat"])
  
  def evaluate(self, time):
    phase = (time - self["T0"])/self["per"]
    phase -= numpy.floor(phase)
    diskBright = 0.5 * ( 1.0 + numpy.cos( (180.0 - phase*360.0)/180.0 * numpy.pi ) * \
                 numpy.sin(self["i"]/180.0*numpy.pi) ) * self["brat"]
    return diskBright 
    


import sys
from pathlib import Path
    
try:
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        import tomli as tomllib
except ImportError:
    tomllib = None


def _extract_version(ll):
    """ Finds and returns version tag using simplistic parsing """
    for l in ll:
        s = l.split("=")
        if s[0].strip() == "version":
            return s[1].strip()
    return "0.0.0"

def _get_version_from_toml():
    """
    Finds and robustly parses pyproject.toml to get the version.
    Returns "0.0.0-unknown" on failure.
    """
    pyproject_path = Path(__file__).resolve().parent.parent / "pyproject.toml"
    if not tomllib:
        # Try simple parse
        with open(pyproject_path, 'rt') as to:
            return _extract_version(to.readlines())
    try:
        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)

        return pyproject_data["project"]["version"]

    except (FileNotFoundError, KeyError, tomllib.TOMLDecodeError) as e:
        print(e)
        return "0.0.0-toml-error"
    
    
# Try to get the metadata library
try:
    if sys.version_info >= (3, 8):
        from importlib import metadata
    else:
        # For older Python, you need the 'importlib-metadata' backport
        import importlib_metadata as metadata
except ImportError:
    # If the backport is not installed, we cannot use this method.
    metadata = None

__version__ = "0.0.0-unknown"

if metadata is not None:
    try:
        __version__ = metadata.version("PyAstronomy")
    except metadata.PackageNotFoundError:
        __version__ = _get_version_from_toml()


def PyA_Version():
    return __version__




# -*- coding: utf-8 -*-
from __future__ import division
import numpy
import PyAstronomy.funcFit as fuf
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.modelSuite.XTran import _ZList


# Check whether mpmath and/or the elliptical integrals from the boost library
# can be imported.

mpmath_imported = False
boostEll_imported = False

try:
    import mpmath
    mpmath_imported = True
except ImportError:
    mpmath_imported = False

try:
    import ell
    boostEll_imported = True
except ImportError:
    boostEll_imported = False


class _Case:
    """ 
    *Case-designation* according to Pal 2008 Table 1.

    Parameters
    ----------
    s : int
        Step identifier in Pal 2008 Table 1.
    c : string
        Case identifier in Pal 2008 Table 1.
    """

    def __init__(self, s, c):
        self.step = s
        self.case = c

    def _str_(self):
        return "Step: " + str(self.step) + "  case: " + str(self.case)


class PalLC(_ZList, fuf.OneDFit):
    """
    Calculate and fit analytical transit light-curves using the formulae
    provided by Pal 2008.

    This class uses a **circular planetary orbit**.

    .. note :: The **evaluation of elliptical integrals** is essential
               in calculating the transit model. While both
               the *mpmath* module and the *Boost* libraries implement
               those integrals, it is the Boost library, which
               evaluates them by far more quickly. Yet, the support
               for Boost has to be added manually. 

    *Model parameters*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `linLib` - Linear limb-darkening coefficient.
      - `quadLimb` - Quadratic limb-darkening coefficient.
      - `T0` - Time offset of transit center.
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).

    This class inherits the functionality of funcFit's OneDFit object.
    You can, therefore, set and get the parameter using the brackets:
      e.g., pallc["p"] = 0.12345

    .. warning::
        Time units have to be consistent.
    """

    def __init__(self):
        _ZList.__init__(self, "circular")
        fuf.OneDFit.__init__(
            self, ["p", "a", "i", "linLimb", "quadLimb", "T0", "per", "b"])
        self.freeze(["p", "a", "i", "linLimb", "quadLimb", "T0", "per", "b"])
        self.setRootName("PalCirc08")

        self._zlist = None

        if (not mpmath_imported) and (not boostEll_imported):
            raise(PE.PyARequiredImport("Neither mpmath nor the elliptical integrals from the boost library could be found!",
                                       where="PalLC::__init__", solution="Install mpmath or make Boost library available (more complicated - see documentation)"))
        self.useBoost = False
        if boostEll_imported:
            self.useBoost = True
            self.ell = ell.ell()

    def _selectCases(self):
        """ 
        Determines which 'case' (according to Pal '08, Table 1) must be
        attributed to an individual value of 'z' and returns a
        list of 'Case' class objects.
        """

        result = []
        p = self["p"]
        for z in self._zlist:
            if z == 0.0 and p < 1.0:
                result.append(_Case(1, 'A'))
                continue
            if z <= p - 1.0:
                result.append(_Case(2, 'Ag'))
                continue
            if z < p and z < (1.0 - p):
                result.append(_Case(3, 'B'))
                continue
            if z < p and z == 1.0 - p:
                result.append(_Case(4, 'Bt'))
                continue
            if z < p:
                result.append(_Case(5, 'Bg'))
                continue
            if z == p and z < 1.0 - p:
                result.append(_Case(6, 'C'))
                continue
            if z == p and z == 0.5:
                result.append(_Case(7, 'Ct'))
                continue
            if z == p:
                result.append(_Case(8, 'Cg'))
                continue
            if z < 1.0 - p:
                result.append(_Case(9, 'D'))
                continue
            if z == (1.0 - p):
                result.append(_Case(10, 'E'))
                continue
            if z < 1.0 + p:
                result.append(_Case(11, 'F'))
                continue
            # This means that there is no occultation at all
            result.append(_Case(12, 'G'))
        return result

    def _ci(self, z, a):
        return 2 / (9.0 * numpy.pi * numpy.sqrt(1.0 - a))

    def _cik(self, z, a, b):
        return 1.0 - 5.0 * z**2.0 + self["p"]**2.0 + a * b

    def _cie(self, z, a):
        return (z**2.0 + 7 * self["p"]**2.0 - 4.0) * (1.0 - a)

    def _cipi(self, z):
        return -3.0 * (self["p"] + z) / (self["p"] - z)

    def _cg(self, z):
        return 1.0 / (9.0 * numpy.pi * numpy.sqrt(self["p"] * z))

    def _cgk(self, z):
        return 3.0 - 6.0 * (1.0 - self["p"] * self["p"])**2.0 - 2.0 * self["p"] * z * (z * z + 7 * self["p"] * self["p"] - 4.0 + 5.0 * self["p"] * z)

    def _cge(self, z):
        return 4.0 * self["p"] * z * (z * z + 7.0 * self["p"] * self["p"] - 4.0)

    def _cgpi(self, z):
        return -3.0 * (self["p"] + z) / (self["p"] - z)

    def _ti(self, z):
        phat = numpy.sqrt(self["p"] * (1.0 - self["p"]))
        return 2.0 / (3.0 * numpy.pi) * numpy.arccos(1.0 - 2.0 * self["p"]) - 4.0 / (9.0 * numpy.pi) * (3.0 + 2.0 * self["p"] - 8.0 * self["p"]**2.0) * phat

    def _g0(self, z):
        k0 = numpy.arccos(
            (self["p"]**2.0 + z**2.0 - 1.0) / (2.0 * self["p"] * z))
        k1 = numpy.arccos((z**2.0 + 1.0 - self["p"]**2.0) / (2.0 * z))
        return (self["p"] * self["p"] * k0 + k1 - numpy.sqrt(z * z - 0.25 * (1 + z * z - self["p"] * self["p"])**2.0)) / numpy.pi

    def _g2(self, z, a, b):
        k0 = numpy.arccos(
            (self["p"]**2.0 + z**2.0 - 1.0) / (2.0 * self["p"] * z))
        k1 = numpy.arccos((z**2.0 + 1.0 - self["p"]**2.0) / (2.0 * z))
        return (k1 + self["p"] * self["p"] * (self["p"] * self["p"] + 2.0 * z * z) * k0 - 0.25 * (1.0 + 5.0 * self["p"] * self["p"] + z * z) * numpy.sqrt((1.0 - a) * (b - 1.0))) / (2.0 * numpy.pi)

    def _returnCoeff(self, step, z):
        a = (self["p"] - z) * (self["p"] - z)
        b = (self["p"] + z) * (self["p"] + z)
        if (step == 1):
            pprime = numpy.sqrt(1.0 - self["p"]**2.0)
            f0 = self["p"]**2.0
            f1 = 2.0 / 3.0 * (1.0 - pprime**3.0)
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 0.5 * self["p"]**4.0
            k = 0.0
            n = 0.0
        elif (step == 2):
            f0 = 1.0
            f1 = 2.0 / 3.0
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 0.5
            k = 0.0
            n = 0.0
        elif (step == 3):
            f0 = self["p"]**2.0
            f1 = 2.0 / 3.0
            fk = self._ci(z, a) * self._cik(z, a, b)
            fe = self._ci(z, a) * self._cie(z, a)
            fpi = self._ci(z, a) * self._cipi(z)
            f2 = 0.5 * self["p"]**2.0 * (self["p"]**2.0 + 2.0 * z**2.0)
            k = numpy.sqrt((4.0 * self["p"] * z) / (1.0 - a))
            n = -4.0 * self["p"] * z / a
        elif (step == 4):
            f0 = self["p"]**2.0
            f1 = self._ti(z)
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 0.5 * self["p"]**2.0 * (self["p"]**2.0 + 2.0 * z**2.0)
            k = 0.0
            n = 0.0
        elif (step == 5):
            f0 = self._g0(z)
            f1 = 2.0 / 3.0
            fk = self._cg(z) * self._cgk(z)
            fe = self._cg(z) * self._cge(z)
            fpi = self._cg(z) * self._cgpi(z)
            f2 = self._g2(z, a, b)
            k = numpy.sqrt((1.0 - a) / (4.0 * self["p"] * z))
            n = (a - 1.0) / a
        elif (step == 6):
            f0 = self["p"]**2.0
            f1 = 1.0 / 3.0
            fk = 2.0 / (9.0 * numpy.pi) * (1.0 - 4.0 * self["p"]**2.0)
            fe = 8.0 / (9.0 * numpy.pi) * (2.0 * self["p"]**2.0 - 1.0)
            fpi = 0.0
            f2 = 3.0 / 2.0 * self["p"]**4.0
            k = 2 * self["p"]
            n = 0.0
        elif (step == 7):
            f0 = 0.25
            f1 = 1.0 / 3.0 - 4.0 / (9.0 * numpy.pi)
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 3.0 / 32.0
            k = 0.0
            n = 0.0
        elif (step == 8):
            f0 = self._g0(z)
            f1 = 1.0 / 3.0
            fk = -1.0 / (9.0 * numpy.pi * self["p"]) * (1.0 -
                                                        4.0 * self["p"]**2.0) * (3.0 - 8.0 * self["p"]**2.0)
            fe = 1.0 / (9.0 * numpy.pi) * 16.0 * \
                self["p"] * (2.0 * self["p"]**2.0 - 1.0)
            fpi = 0.0
            f2 = self._g2(z, a, b)
            k = 1.0 / (2.0 * numpy.pi)
            n = 0.0
        elif (step == 9):
            f0 = self["p"]**2.0
            f1 = 0.0
            fk = self._ci(z, a) * self._cik(z, a, b)
            fe = self._ci(z, a) * self._cie(z, a)
            fpi = self._ci(z, a) * self._cipi(z)
            f2 = 0.5 * self["p"]**2.0 * (self["p"]**2.0 + 2.0 * z**2.0)
            k = numpy.sqrt((4.0 * self["p"] * z) / (1.0 - a))
            n = -4.0 * self["p"] * z / a
        elif (step == 10):
            f0 = self["p"]**2.0
            f1 = self._ti(z)
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 0.5 * self["p"]**2.0 * (self["p"]**2.0 + 2.0 * z**2.0)
            k = 0.0
            n = 0.0
        elif (step == 11):
            f0 = self._g0(z)
            f1 = 0.0
            fk = self._cg(z) * self._cgk(z)
            fe = self._cg(z) * self._cge(z)
            fpi = self._cg(z) * self._cgpi(z)
            f2 = self._g2(z, a, b)
            k = numpy.sqrt((1.0 - a) / (4.0 * self["p"] * z))
            n = (a - 1.0) / a
        elif (step == 12):
            f0 = 0.0
            f1 = 0.0
            fk = 0.0
            fe = 0.0
            fpi = 0.0
            f2 = 0.0
            k = 0.0
            n = 0.0
        return [f0, f1, fk, fe, fpi, f2, float(k), float(n)]

    def whichEllInts(self):
        """
        Says which module is used for the evaluation of the elliptical functions.

        .. note::
           The support for BOOST has to be added manually!

        Returns
        -------
        Identifier : string
            Either 'boost' or 'mpmath' depending on whether the *BOOST*
            libraries are used or the *mpmath* python module.
        """
        if self.useBoost:
            return "boost"
        return "mpmath"

    def evaluate(self, time):
        """ 
        Calculate a light curve according to the analytical models
        given by Pal 2008.

        Parameters
        ----------
        time : array
            An array of time points at which the light curve
            shall be calculated.

        .. note:: time = 0 -> Planet is exactly in the line of sight (phase = 0).

        Returns
        -------
        Model : array
            The analytical light curve is stored in the property `lightcurve`.
        """

        # Translate the given parameters into an orbit and, finally,
        # into a projected, normalized distance (z-parameter)
        self._calcZList(time - self["T0"])

        # 'W' parameters corresponding to notation in Pal '08
        w = 6. - 2. * self["linLimb"] - self["quadLimb"]
        w0 = (6. - 6. * self["linLimb"] - 12. * self["quadLimb"]) / w
        w1 = (6. * self["linLimb"] + 12. * self["quadLimb"]) / w
        w2 = 6. * self["quadLimb"] / w

        # Initialize flux decrease array
        df = numpy.zeros(len(time))
        # Get a list of 'cases' (according to Pal '08). Depends on radius ratio and 'z'-parameter along the orbit
        ca = self._selectCases()
        # Loop through z-list, and calculate the light curve at each point in z (->time)
        for i in self._intrans:
            # Calculate the coefficients to be substituted into the Pal '08 equation
            c = self._returnCoeff(ca[i].step, self._zlist[i])
            # Substitute the coefficients and get 'flux decrease'
            if ca[i].step != 12:
                # Calculate flux decrease only if there is an occultation
                if not self.useBoost:
                    df[i] = w0 * c[0] + w2 * c[5] + w1 * (c[1] + c[2] * mpmath.ellipk(
                        c[6]**2) + c[3] * mpmath.ellipe(c[6]**2) + c[4] * mpmath.ellippi(c[7], c[6]**2))
                else:
                    df[i] = w0 * c[0] + w2 * c[5] + w1 * (c[1] + c[2] * self.ell.ell1(
                        c[6]) + c[3] * self.ell.ell2(c[6]) + c[4] * self.ell.ell3(c[7], c[6]))
        self.lightcurve = (1. - df) * 1. / \
            (1. + self["b"]) + self["b"] / (1.0 + self["b"])
        return self.lightcurve


class PalLCKep(PalLC):
    """
    Analytical transit light-curves using the formulae provided by Pal 2008.

    More information on the Keplerian orbit can be found here: :ref:`keplerorbitpyasl`

    .. note :: The **evaluation of elliptical integrals** is essential
               in calculating the transit model. While both
               the *mpmath* module and the *Boost* libraries implement
               those integrals, it is the Boost library, which
               evaluates them by far more quickly. Yet, the support
               for Boost has to be added manually. 

    *Model parameters*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `w` - Longitude of periastron [deg].
      - `Omega` - Longitude of the ascending node [deg].
      - `e` - Orbital eccentricity.
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `linLib` - Linear limb-darkening coefficient.
      - `quadLimb` - Quadratic limb-darkening coefficient.
      - `tau` - Time of periastron passage.
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).

    This class inherits the functionality of funcFit's OneDFit object.
    You can, therefore, set and get the parameter using the brackets:
      e.g., pallc["p"] = 0.12345

    .. warning::
        Time units have to be consistent.

    Parameters
    ----------
    ke : None, optional
        If an instance of KeplerEllipse is given, the parameters a, i, tau, per, w, Omega,
        and e will be adopted from it.
    collisionCheck : boolean, optional
        If set True, it will be checked whether the two bodies collide on the current orbit.
    """

    def __init__(self, ke=None, collisionCheck=False):
        _ZList.__init__(self, "keplerian", collisionCheck)
        fuf.OneDFit.__init__(self, ["p", "a", "i", "linLimb", "quadLimb", "tau", "per", "b",
                                    "w", "Omega", "e"])
        self.freeze(["p", "a", "i", "linLimb", "quadLimb", "tau", "per", "b",
                     "w", "Omega", "e"])
        self.setRootName("Pal08")
        if ke is None:
            self["per"] = 1.0
            self["w"] = -90
        else:
            self["per"] = ke.per
            self["w"] = ke.w
            self["i"] = ke.i
            self["tau"] = ke.tau
            self["a"] = ke.a        
            self["Omega"] = ke.Omega
            self["e"] = ke.e    

        self._zlist = None

        if (not mpmath_imported) and (not boostEll_imported):
            raise(PE.PyARequiredImport("Neither mpmath nor the elliptical integrals from the boost library could be found!",
                                       where="PalLC::__init__", solution="Install mpmath or make Boost library available (more complicated - see documentation)"))
        self.useBoost = False
        if boostEll_imported:
            self.useBoost = True
            self.ell = ell.ell()

        self.collisionCheck = collisionCheck

        # Wrap get/setitem to inform about change in parameter name
        T0paE = PE.PyADeprecationError("The parameter 'T0pa' in PalLCKep had to be renamed 'tau'.",
                                       solution="Use 'tau' instead of 'T0pa'.")

        def getitem(specifier, **kwargs):
            if specifier == "T0pa":
                PE.warn(T0paE)
            return PalLC.__getitem__(self, specifier, **kwargs)
        self.__getitem__ = getitem

        def setitem(specifier, value):
            if specifier == "T0pa":
                PE.warn(T0paE)
            PalLC.__setitem__(self, specifier, value)
        self.__setitem__ = setitem

    def evaluate(self, time):
        """ 
        Calculate a light curve according to the analytical models
        given by Pal 2008.

        Parameters
        ----------
        time : array
            An array of time points at which the light curve
            shall be calculated.

        Returns
        -------
        Model : array
            The analytical light curve is stored in the property `lightcurve`.
        """

        # Translate the given parameters into an orbit and, finally,
        # into a projected, normalized distance (z-parameter)
        self._calcZList(time)

        # 'W' parameters corresponding to notation in Pal '08
        w = 6. - 2. * self["linLimb"] - self["quadLimb"]
        w0 = (6. - 6. * self["linLimb"] - 12. * self["quadLimb"]) / w
        w1 = (6. * self["linLimb"] + 12. * self["quadLimb"]) / w
        w2 = 6. * self["quadLimb"] / w

        # Initialize flux decrease array
        df = numpy.zeros(len(time))
        # Get a list of 'cases' (according to Pal '08). Depends on radius ratio and 'z'-parameter along the orbit
        ca = self._selectCases()
        # Loop through in-transit points in z-list,
        # and calculate the light curve at each point in z (->time)
        for i in self._intrans:
            # Calculate the coefficients to be substituted into the Pal '08 equation
            c = self._returnCoeff(ca[i].step, self._zlist[i])
            # Substitute the coefficients and get 'flux decrease'
            if ca[i].step != 12:
                # Calculate flux decrease only if there is an occultation
                if not self.useBoost:
                    df[i] = w0 * c[0] + w2 * c[5] + w1 * (c[1] + c[2] * mpmath.ellipk(
                        c[6]**2) + c[3] * mpmath.ellipe(c[6]**2) + c[4] * mpmath.ellippi(c[7], c[6]**2))
                else:
                    df[i] = w0 * c[0] + w2 * c[5] + w1 * (c[1] + c[2] * self.ell.ell1(
                        c[6]) + c[3] * self.ell.ell2(c[6]) + c[4] * self.ell.ell3(c[7], c[6]))
        self.lightcurve = (1. - df) * 1. / \
            (1. + self["b"]) + self["b"] / (1.0 + self["b"])
        return self.lightcurve


PalLC_Rebin = fuf.turnIntoRebin(PalLC)




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Gaussian fit and parameter handling
        # ===================================
        
        # In[1]:
        
        
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones_like
        import matplotlib.pylab as plt
        import scipy.optimize as sco
        # ... and now the funcFit2 package
        from PyAstronomy import funcFit2 as fuf2
        import matplotlib
        matplotlib.use('Agg')
        
        # For reproducability
        random.seed(1234)
        
        # Create some mock data
        # Choose parameters...
        gPar = {"A":-5.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # ...and calculate profile...
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2*pi*gPar["sig"]**2) \
            * exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # ...add some noise...
        y += random.normal(0.0, 0.01, x.size)
        # ...and save the error bars
        yerr = ones_like(x)*0.01
        
        plt.errorbar(x, y, yerr=yerr, fmt='b+')
        plt.show()
        
        # Fitting and parameter management
        # ---------------------------------------------------------------
        
        # In[2]:
        
        
        # Create a fitting object representing a Gaussian and set guess parameters.
        gf = fuf2.GaussFit()
        # See what parameters are available
        print("List of available parameters: ", gf.availableParameters())
        print()
        
        # Set guess values for the parameters
        gf["A"] = -10.0
        gf["sig"] = 15.77
        gf["off"] = 0.96
        gf["mu"] = 7.5
        print()
        
        # Let us see whether the assignment worked
        print("Parameters and guess values: ")
        print("  A   : ", gf["A"])
        print("  sig : ", gf["sig"])
        print("  off : ", gf["off"])
        print("  mu  : ", gf["mu"])
        print()
        
        # More convenient overview of parameter status
        gf.parameterSummary()
        print()
        
        # Exporting and assigning parameter values via dictionary
        ps = gf.parameters()
        print("Parameter name/value dictionary: ", ps)
        # Assigning values from dictionary (identical values here)
        gf.assignValues(ps)
        
        
        # Evaluate the model for the starting values
        startmodel = gf.evaluate(x)
        
        # Which parameters shall be varied during the fit?
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["A", "sig", "off", "mu"])
        
        # Show values and names of free/frozen parameters (i.e., parameters not varied in a fit)
        print("Names and values of FREE parameters: ", gf.freeParameters())
        print("Names and values of FROZEN parameters: ", gf.frozenParameters())
        print()
        
        # Use scipy's fmin to minimize chi square 
        fr = sco.fmin(gf.chisqr, gf.freeParamVals(), args=(x,y,yerr), full_output=True)
        print()
        print("Fit results: ", fr)
        # Set the parameter values to best-fit values
        gf.setFreeParamVals(fr[0])
        
        
        # Use a convenience function to carry out optimization
        # using scipy's fmin algorithm (parameter values are set to
        # best-fit result)
        fr = fuf2.fitfmin(gf, gf.chisqr, x, y, yerr=yerr)
        
        # Get a summary of current parameters from the model
        gf.parameterSummary()
        
        # Let us see what we have done...
        bestfitmodel = gf.evaluate(x)
        
        plt.plot(x, y, 'bp')
        plt.plot(x, startmodel, 'g:', label="Model for starting values")
        plt.plot(x, bestfitmodel, 'r--', label="Best-fit model")
        plt.legend()
        plt.show()
        



from __future__ import print_function, division
import numpy
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo

class Scanner:
  
  def __init__(self, minVal=0.1, maxVal=10.0, dVal=0.05, mode="period"):
    """
      The Scanner class is used to define the period/frequency range and steps used in the \
      PDM analysis. It is iteratable.
      
      Parameters:
        - `minVal` - float, Minimum value,
        - `maxVal` - float, Maximum value,
        - `dVal`   - float, Delta value,
        - `mode`   - string, optional, Either "period" or "frequency" (default = "period").
      
      .. Note: Whether min, max, and dval refer to period or frequency depends on the ``mode'' parameter.
      .. Note: Time units have match the time axis given to the PDM analysis class.
    """
    if (mode != "period") and (mode != "frequency"):
      raise(PE.PyAValError("`mode` must either be 'period' or 'frequency'. Current value is: "+str(mode), \
                           where="Scanner::__init__", solution="Choose either 'period' or 'frequency' for mode parameter."))
    self.minVal = minVal
    self.maxVal = maxVal
    self.dVal = dVal
    self.mode = mode
  
  def __iter__(self):
    self.curVal = self.minVal
    while self.curVal <= self.maxVal:
      yield self.curVal
      self.curVal += self.dVal



class PyPDM:
  
  def __init__(self, time, mag):
    """
      This class allows to carry out a ``Phase Dispersion Minimization'' (PDM) analysis as described in
      Stellingwerf 1978.
      
      Parameters:
       - time - array, The time array (arbitrary time units),
       - mag - array, The corresponding flux in magnitudes. 
      
      Properties:
       - minBinPoints - The minimum number of data points,
                        which may be contained in an individual bin (default = 3).
    """
    # The minimum number of data points per bin.
    self.minBinPoints = 3
    # Save the data
    self.x = time.copy()
    self.y = mag.copy()
  
  def __setUpEquiBlocks(self, nbins, phases):
    """
      Set up a sequence of equidistant bins. Then check whether all bins contain enough data points \
      (defined by minBinPoints property). \
      If this is not the case, the border to the adjacent bin with fewer photons is dropped. Therefore,
      the resulting bins are not necessarily equidistant anymore.
      
      Parameters:
       - nbins - int, The number of bins.
       - phases - array, The phases of the data points. 
      
      Returns: A list with the beginnings and one list with the ends of the bins (in phase). 
    """
    blockBegin = (numpy.arange(nbins) / float(nbins)).tolist()
    blockEnd   = ((numpy.arange(nbins) + 1.0) / float(nbins)).tolist()
    Ns = []
    for i in smo.range(nbins):
      indi = numpy.where(numpy.logical_and(phases >= blockBegin[i], phases < blockEnd[i]))[0]
      Ns.append(len(indi))
    
    i = 0
    while(i < len(Ns)):
      if Ns[i] < self.minBinPoints:
        # The bin with number i does not contain enough data points.
        # Remove the border to make the bins larger...
        iPlus = i + 1
        iMinu = i - 1
        NPlus = 0; NMinu = 0
        if iPlus == len(blockEnd):
          NPlus = 1e100
        else:
          NPlus = Ns[iPlus]
        if iMinu == -1:
          NMinu = 1e100
        else:
          NMinu = Ns[iMinu]
        if NMinu <= NPlus:
          # Eliminate the border to bin below
          blockBegin.pop(i); blockEnd.pop(i-1)
          Ns[i] = Ns[i-1] + Ns[i]; Ns.pop(i-1)
          # Reiterate
          i = 0
          continue
        else:
          # Eliminate the border to bin behind
          blockBegin.pop(i+1); blockEnd.pop(i)
          Ns[i+1] = Ns[i] + Ns[i+1]; Ns.pop(i)
          # Reiterate
          i = 0
          continue
      i += 1
    return blockBegin, blockEnd

  def __setUpEquiBlocksCover(self, nbins, phases, covers):
    """
      Set up a sequence of equidistant bins. Then reproduce the same sequence but offset by \
      1/(nbins*covers). In this vein, data points are covered by multiple bins. The phase \
      axis is cyclic. This function pays no attention to the number of data points within \
      individual bins.
      
      Parameters:
       - nbins - int, The number of bins.
       - phases - array, The phases of the data points. 
       - covers - int, The number of covers (as explained above).
      
      Returns: A list with the beginnings and one list with the ends of the bins (in phase). 
    """
    blockBegin = numpy.arange(nbins) / float(nbins)
    blockEnd   = (numpy.arange(nbins) + 1.0) / float(nbins)
    
    blockBeginOrig = blockBegin.copy()
    blockEndOrig = blockEnd.copy()
    
    offset = 0
    for i in smo.range(1,covers):
      offset = float(i)/float(nbins * covers)
      blockBegin = numpy.append(blockBegin, blockBeginOrig + offset)
      blockEnd   = numpy.append(blockEnd, blockEndOrig + offset)
      
    return blockBegin, blockEnd

  def phase(self, time, period):
    """
      Calculate phases and ordering.
    
      Parameter:
        - `time` - array, Time stamps,
        - `period` - float, The period used to calculate phases.
    
      Returns:
        An unsorted phase array and an array with indices specifying the order. Thus,
        phase[indi], mag[indi] is the phase-sorted phased light curve.
    """
    phase = time/period - numpy.floor(time/period)
    indi = numpy.argsort(phase)
    return phase, indi
  
  def __getThet(self, phase, mag, bbegin, bend):
    """
      Calculate the Theta statistics defined by Stellinger '78.
      
      Parameters:
       - phase, mag - arrays, Phase-sorted(!) light curve.
       - bbegin, bend -lists, begin and end of individual bins (``blocks'').
      
      Returns the value of Theta.
    """
    meanMag = mag.mean()
    N = len(mag)
    M = len(bbegin)
    sigmaSqr = ((mag - meanMag)**2).sum() / float(N - 1)

    sSqrUp = 0.0; sSqrDown = 0.0
    for i in smo.range(len(bbegin)):
      # Points belonging to a chunk
      indi = numpy.where(numpy.logical_and(phase >= bbegin[i], phase < bend[i]))[0]
      nj = float(len(indi))
      if nj > 1:
        # Variance of individual block
        sigmaj = ((mag[indi] - mag[indi].mean())**2).sum() / float(nj - 1)
        sSqrUp += (nj - 1)*sigmaj
        sSqrDown += nj
    sSqrDown -= M
    
    sSqr = sSqrUp / sSqrDown
    theta = sSqr / sigmaSqr
    return theta
  
  def pdmEquiBin(self, nbins, scanner):
    """
      Carry out the PDM analysis using equidistant bins.
      
      This method uses equidistant bins, yet, it pays attention to the number of \
      data points contained in individual bins. If this number is insufficient \
      (as defined by the `minBinPoints` property), adjacent bins are combined
      (see documentation of *__setUpEquiBlocks*).
      
      Parameters:
        - `nbins` - int, Number of bins to use.
        - `scanner` - An instance of the *Scanner* class, defining, which periods/frequencies
                      are tested.
      
      Returns two arrays: the periods and the associated values of the Theta statistic. These
      are also saved to the class properties `periods`, `frequencies`, and `theta`.

      .. Note:: Whether the first return value is frequency or period
                depends on the mode of the scanner.
    """
    # Go through periods
    periods = []; theta = []
    for val in scanner:
      if scanner.mode == "period":
        period = val
      else:
        period = 1.0/val
      # Note the p[indi], self.y[indi] is the ordered, phased light curve
      p, indi = self.phase(self.x, period)
      bbegin, bend = self.__setUpEquiBlocks(nbins, p[indi])
      periods.append(period)
      theta.append(self.__getThet(p[indi], self.y[indi], bbegin, bend))
    
    # Save result
    self.periods = numpy.array(periods)
    self.frequencies = 1.0 / numpy.array(periods)
    self.theta = numpy.array(theta)
    
    if scanner.mode == "period":
      return self.periods.copy(), self.theta.copy()
    else:
      return self.frequencies.copy(), self.theta.copy()

  def pdmEquiBinCover(self, nbins, covers, scanner):
    """
      Carry out the PDM analysis using multiple sequences of equidistant bins.
      
      The bins used by this method are equidistant, but the phase axis and, thus, \
      the data points are covered by multiple bins. The first sequence of bins is \
      the usual devision of the 0-1 interval into *nbins* bins; the following sequences \
      are the same but offset of 1/(nbins*covers) in phase (the phase axis is cyclic).
      
      This method does not check whether bins contain "enough" data points, i.e., it neglects \
      the `minBinPoints` property taken into account by *pdmEquiBin*. If, however, less than \
      two data points are contained within a bin, it is neglected.
      
      Parameters:
        - `nbins` - int, Number of bins to use.
        - `scanner` - An instance of the *Scanner* class, defining, which periods/frequencies are tested.
        - `covers` - int, The number of covers, i.e., phase-shifted bin sets.
      
      Returns two arrays: the periods/frequencies and the associated values of the Theta statistic. These
      are also saved to the class properties ``periods'', ``frequencies'', and ``theta''.
      
      .. Note:: Whether the first return value is frequency or period depends on the mode of the scanner.
    """
    # Go through periods
    periods = []; theta = []
    for val in scanner:
      if scanner.mode == "period":
        period = val
      else:
        period = 1.0/val
      # Note the p[indi], self.y[indi] is the ordered, phased light curve
      p, indi = self.phase(self.x, period)
      # Extend phase and mag arrays to allow for phases > 1.0 (cyclic phase array...)
      phase = numpy.append(p[indi], p[indi]+1.0)
      mag = numpy.append(self.y[indi], self.y[indi])
      bbegin, bend = self.__setUpEquiBlocksCover(nbins, phase, covers)
      periods.append(period)
      theta.append(self.__getThet(phase, mag, bbegin, bend))
    
    # Save result
    self.periods = numpy.array(periods)
    self.frequencies = 1.0 / numpy.array(periods)
    self.theta = numpy.array(theta)
    
    if scanner.mode == "period":
      return self.periods.copy(), self.theta.copy()
    else:
      return self.frequencies.copy(), self.theta.copy()
    
    
    
    


from __future__ import print_function, division
import unittest


class TestSanityOfPyaslExt1(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_ExampleTransmissionCurves(self):
        """
        Checking example for 'TransmissionCurves'
        """
        # from __future__ import print_function
        from PyAstronomy import pyasl
        import numpy as np
        import matplotlib.pylab as plt
        # Import six for Python 2/3 compatibility
        import six

        # Get transmission curve object
        tcs = pyasl.TransmissionCurves()

        # Add passbands from Spitzer IRAC
        tcs.addSpitzerIRACPassbands()

        print("Available bands: ", tcs.availableBands())

        # Wavelength axis
        wvl = np.linspace(3000, 10000, 10000)

        # Plot transmission curves for Bessel b, v, and r bands
        for (b, c) in six.iteritems({"b": "b", "v": "k", "r": "r"}):
            tc = tcs.getTransCurve("Bessel " + b)
            trans = tc(wvl)
            plt.plot(wvl, trans, c+'-', label="Bessel " + b)

        # Plot transmission curves for Johnson U, B, and V bands
        for (b, c) in six.iteritems({"U": "m", "B": "b", "V": "k"}):
            tc = tcs.getTransCurve("Johnson " + b)
            trans = tc(wvl)
            plt.plot(wvl, trans, c+'--', label="Johnson " + b)

        plt.legend()
        plt.xlabel(r"Wavelength [$\AA$]")
        plt.ylabel("Transmission")
        # plt.show()

        # Create Planck spectrum ...
        wvl = np.arange(3000., 10000., 1.0)
        spec = pyasl.planck(T=5777., lam=wvl*1e-10)
        # ... and convolve with Johnson V band
        vbs = tcs.convolveWith(wvl, spec, "Johnson V")

        plt.plot(wvl, spec, 'b-', label='Input spectrum')
        plt.plot(wvl, vbs, 'r--', label='Convolution with Johnson V band')
        plt.legend()
        # plt.show()
        
    def testsanity_DownloadTransmissionCurves(self):
        """
        Checking downloads of 'TransmissionCurves'
        """
        from PyAstronomy import pyasl
        tcs = pyasl.TransmissionCurves()
        tcs.addSpitzerIRACPassbands(forceDownload=True)
        tcs.addKeplerPassband(forceDownload=True)
        tcs.addTESSPassband(forceDownload=True)



# -*- coding: utf-8 -*-
from __future__ import division
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np


def smooth(x, windowLen, window='flat'):
    """
    Smooth data using a window function.

    This method is based on the convolution of a window function with the signal.
    The window function is normalized so that the sum of its entries amounts to
    one. The signal is prepared by adding reflected copies of the signal 
    (with the window size) to both ends of the input array, so that the output
    array can have the same length as the input. Consequently the smoothing at
    the edges is actually based on extrapolation.

    .. note:: This algorithm was adopted from the scipy cookbook
              (http://www.scipy.org/Cookbook/SignalSmooth). The copyright
              of the original algorithm belongs to the authors of that
              cookbook algorithm.

    Parameters
    ----------
    x : array
        The input signal 
    windowLen : int
        The dimension of the smoothing window. It must be an odd integer.
    window : string, {'flat', 'hanning', 'hamming', 'bartlett', 'blackman'}
        The window function to be used. A flat window will
        produce a moving average smoothing.

    Returns
    -------
    Smoothed signal : array
        The smoothed signal. Same length as input array.
    """

    if x.ndim != 1:
        raise(PE.PyAValError("Only one dimensional arrays can be smoothed. Dimension of " +
                             "given array is " + str(x.ndim),
                             solution="Change dimension of input."))

    if x.size < windowLen:
        raise(PE.PyAValError("Input vector needs to be bigger than window size.",
                             solution="Check the length of the input array and window size."))

    if windowLen < 3:
        PE.warn(PE.PyAValError("Length of window is smaller then 3. No smoothing is done.",
                               solution="Check window size."))
        return x

    if windowLen % 2 != 1:
        raise(PE.PyAValError("Parameter `windowLen` should be an odd integer"))

    # Extend input array at the edges to have the same
    # length for the output array. Insert a mirrored version
    # of the first part of the data array in front of the
    # first data point; apply the same scheme to the end of the
    # data array.
    s = np.r_[x[windowLen - 1:0:-1], x, x[-1:-windowLen:-1]]

    if window == 'flat':
        # This is a moving average
        w = np.ones(windowLen, 'd')
    elif window == "hanning":
        w = np.hanning(windowLen)
    elif window == "hamming":
        w = np.hamming(windowLen)
    elif window == "bartlett":
        w = np.bartlett(windowLen)
    elif window == "blackman":
        w = np.blackman(windowLen)
    else:
        raise(PE.PyAValError("Current `window` parameter (" + str(window) + ") is not supported. " +
                             "Must be one of: 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'",
                             solution="Choose one of the above window types."))

    y = np.convolve(w / w.sum(), s, mode='valid')
    return y[(windowLen // 2):-(windowLen // 2)]



import numpy as np
from PyAstronomy import funcFit as fuf
from PyAstronomy import pyasl

class RotBroadProfile(fuf.OneDFit):
  """
    Implements rotational broadening with linear limb-darkening.
    
    *Fit Parameters*:
      - `xmax`  - Maximal extent of the profile
      - `eps`   - Linear limb-darkening coefficient
      - `A`     - Area under the profile (negative for absorption)
      - `off`   - An offset applied to the profile
      - `lin`   - Gradient of a linear term to adjust the 'continuum'
      - `mu`    - Center of the profile (same units as `xmax`)
      - `gsig`  - The standard deviation of a Gaussian with which the
                  rotational profile is convoluted, e.g., to model
                  instrumental resolution.

    The profile is given by:
    
    .. math::
        
        G(x) = A \\left(c_1\\sqrt(1-(x/x_{max})^2) + c_2(1-(x/x_{max})^2)\\right) + off + lin \\cdot x
    
    with the constants given by:
    
    .. math::
    
        c_1 = \\frac{2(1-\\epsilon)}{\\pi x_{max} (1-\\epsilon/3)} \\;\\;\\;\\;
        c_2 = \\frac{\\epsilon}{2 x_{max} (1-\\epsilon/3)}
    
    Here, `x` can either denote a velocity or a wavelength shift. Thus, `xmax` can
    be given in km/s or Angstrom depending on the input; see, e.g., "Stellar Photospheres"
    by D.F. Gray for a derivation.
    
    Note that the profile is normalized, i.e., the area amounts to one. This may, however,
    depend on the sampling. If the profile is undersampled, errors can become significant.
  """

  def __init__(self):
    fuf.OneDFit.__init__(self, ["off", "lin", "xmax", "eps", "A", "mu", "gsig"], rootName="RTB")
    self["A"] = 1.0
    self["xmax"] = 1.0
  
  def _rotProf(self, v):
    """
      Calculate the rotationally broadened line-profile.
      
      Parameters
      ----------
      v : array
          Abscissa at which to evaluate profile.
      
      Returns
      -------
      Profile : array
          The rotationally broadened profile.
    """
    denom = np.pi*self["xmax"]*(1.0-self["eps"]/3.0)
    c1 = 2.0*(1.0-self["eps"])/denom
    c2 = (np.pi*self["eps"]/2.0) / denom
    
    indi = np.where(self["xmax"] - np.abs(v-self["mu"]) > 0.0)[0]
    y = np.zeros(v.size)
    
    vv = ((v[indi]-self["mu"])/self["xmax"])**2
    y[indi] += (c1 * np.sqrt(1.0 - vv) + c2 * (1.0 - vv))
    y[indi] *= self["A"]
    return y
  
  def evaluate(self, v):
    """
      Calculates the rotational broadening profile according to current parameter values.

      Parameters:
      x : array
          Wavelength or velocity.
    """
    y = self._rotProf(v)
    if self["gsig"] != 0.0:
      # Convolve with Gaussian to implement instrumental broadening
      y = pyasl.broadGaussFast(v, y, self["gsig"], edgeHandling="firstlast")
    y += self["off"] + (self["lin"] * v)
    return y


import numpy as np
from PyAstronomy import pyasl
from PyAstronomy.pyaC import pyaErrors as PE


def solarDirectFluxMeinel(z, height=0.0, I0=1353, c=0.357, s=0.678, a=0.14):
    r"""
    Direct (unscattered) solar flux in the Earth's atmosphere according to Meinel 1976.

    The height-dependent flux of direct solar radiation was given by
    Meinel & Meinel `` Applied Solar Energy'' (Addison Wesley Publishing Co., 1976)
    on page 46 as

    .. math::

        I_0 \left( (1-ah) \exp(-c \times A(z)^s) + ah  \right)

    where :math:`A(z)` is the airmass (evaluated via :func:`airmassPP` using
    the plane parallel approximation)

    Parameters
    ----------
    z : float or array
        The zenith angle [deg]
    height : float
        The height above ground [km]
    I0 : float, optional
        Solar constant (W/m**2)
    c : float, optional
        Coefficient
    s : float, optional
        Coefficient
    a : float, optional
        Coefficient

    Returns
    -------
    Direct flux : float or array
        Direct solar flux in W/m**2 facing the Sun

    """
    f1 = np.exp(-c)
    am = pyasl.airmassPP(z)
    if height == 0.0:
        return I0 * f1**(am**s)
    else:
        ah = a*height
        if ah >= 1:
            raise (PE.PyAValError(
                "Height is too large for the calculation: (1-ah) is negative."))
        return I0*((1-ah)*f1**(am**s) + ah)



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
import PyAstronomy.pyaC as pyaC
import os
import gzip
import csv
import numpy as np
import six.moves as smo
import six


class ExoplanetsOrg(pp.PyAUpdateCycle):
    """
    Easy access to explanets.org exoplanet archive.

    This class downloads a table of exoplanet data from
    the explanets.org
    (http://exoplanets.org/csv-files/)
    and provides access to these data. By default, the
    data will be re-downloaded every seven days.

    The following data are provided:

    ===========   ===================================  ======
    Column        Description                          Unit
    -----------   -----------------------------------  ------
    pl_name       Name of the planet
    pl_orbper     Planetary orbital period             d
    pl_massj      Planetary mass                       MJ
    pl_msini      Minimum planetary mass               MJ
    pl_radj       Planetary radius                     RJ
    pl_trandep    Central depth of transit             %
    pl_impact     Impact Parameter                     Stellar Radii
    pl_trandur    Transit duration                     d
    pl_tranmid    Transit midpoint                     BJD
    pl_tperi      Time of Periastron passage           BJD
    pl_orbsmax    Semi-major-axis                      AU
    pl_orbsmaxr   Semi-major-axis / R_star             Stellar Radii
    pl_orbincl    Orbital inclination of planet        deg
    pl_missal     Orbital misalignment of planet       deg
    pl_omega      Argument of Periastron               deg
    pl_ecc        Planetary orbital eccentricity       %
    pl_grav       Planetary surface gravity            log10(cm/s^2)
    pl_dens       Planetary Density                    g/cm^3
    pl_dtype      Detection type                       %
    KOI           Kepler ID (if available)

    pl_hostname   Name of host star
    st_binary     Binary Flag
    st_rad        Stellar radii                        Solar
    st_dist       Distance to star                     pc
    st_par        Stellar parallax                     mas
    st_mass       Stellar mass                         Solar
    st_teff       Effective temperature of star        K
    st_vsini      Stellar vsin(i)                      km/s
    st_logg       Stellar surface gravity              cm/s**2
    st_acts       Stellar S-Index
    st_actc       Stellar chromospheric activity
    st_vj         Stellar V-band brightness            mag
    st_fe         Stellar metallicity
    st_radv       System radial velocity               km/s
    st_dens       Density of star                      g/cm**3
    K             Velocity Semi-amplitude              m/s

    dec           Declination (J2000)                  dms
    ra            Right ascension (J2000)              hms
    ===========   ===================================  ======
    """

    def _downloadData(self):
        """
        Download data.
        """
        url = "http://exoplanets.org/csv-files/exoplanets.csv"
        self._fs.downloadToFile(
            url, self.dataFileName, clobber=True, verbose=False, openMethod=gzip.open
        )

    def __init__(self, skipUpdate=False, verbose=False):
        self.data = None
        self._verbose = verbose
        self.dataFileName = os.path.join("pyasl", "resBased", "eporg.csv.gz")
        configFilename = os.path.join("pyasl", "resBased", "eporg.cfg")
        pp.PyAUpdateCycle.__init__(self, configFilename, "eporgupdate")
        # Define columns to select
        # Column name, Description, Unit
        self._columns = {}
        # planetary parameters
        self._columns[0] = ["pl_name", "Name of the planet", "", "U15"]
        self._columns[1] = ["pl_orbper", "Planetary orbital period", "d", float]
        self._columns[2] = ["pl_massj", "Planetary mass", "MJ", float]
        self._columns[3] = ["pl_msini", "Minimum planetary mass", "MJ", float]
        self._columns[4] = ["pl_radj", "Planetary radius", "RJ", float]
        self._columns[5] = [
            "pl_trandep",
            "Central depth of transit",
            "(r_p/r_star)^2",
            float,
        ]
        self._columns[6] = ["pl_impact", "Impact Parameter", "Stellar Radii", float]
        self._columns[7] = ["pl_trandur", "Duration of transit", "d", float]
        self._columns[8] = ["pl_tranmid", "Transit midpoint", "BJD", float]
        self._columns[9] = ["pl_tperi", "Time of Periastron passage", "d", float]
        self._columns[10] = ["pl_orbsmax", "Semi-major-axis", "AU", float]
        self._columns[11] = [
            "pl_orbsmaxr",
            "Ratio sma to R_star",
            "Stellar Radii",
            float,
        ]
        self._columns[12] = [
            "pl_orbincl",
            "Orbital inclination of planet",
            "deg",
            float,
        ]
        self._columns[13] = [
            "pl_missal",
            "Orbital misalignment of planet ",
            "deg",
            float,
        ]
        self._columns[14] = ["pl_omega", "Argument of Periastron", "deg", float]
        self._columns[15] = ["pl_ecc", "Planetary orbital eccentricity", "", float]
        self._columns[16] = [
            "pl_grav",
            "Planetary surface gravity",
            "log10(cm/s^2)",
            float,
        ]
        self._columns[17] = ["pl_dens", "Planetary Density", "g/cm^3", float]
        self._columns[18] = ["pl_dtype", "Detection type", "", "U27"]
        self._columns[19] = ["KOI", "Kepler ID (if available)", "", float]
        # stellar pars
        self._columns[20] = ["pl_hostname", "Name of host star", "", "U15"]
        self._columns[21] = ["st_binary", "Binary Flag", "", float]
        self._columns[22] = ["st_rad", "Radius of star", "Solar", float]
        self._columns[23] = ["st_dist", "Distance to Star", "pc", float]
        self._columns[24] = ["st_par", "Stellar parallax", "mas", float]
        self._columns[25] = ["st_mass", "Stellar mass", "Solar", float]
        self._columns[26] = ["st_teff", "Stellar effective temperature", "K", float]
        self._columns[27] = ["st_vsini", "Stellar vsin(i)", "km/s", float]
        self._columns[28] = ["st_logg", "Stellar surface gravity", "cm/s**2", float]
        self._columns[29] = ["st_acts", "Stellar S-Index ", "", float]
        self._columns[30] = ["st_actc", "Stellar chromospheric activity", "", float]
        self._columns[31] = ["st_vj", "Stellar V-band brightness", "mag", float]
        self._columns[32] = ["st_fe", "Stellar metallicity", "", float]
        self._columns[33] = ["st_radv", "System radial velocity", "km/s", float]
        self._columns[34] = ["st_dens", "Density of star", "g/cm^3", float]
        self._columns[35] = ["K", "Velocity Semi-amplitude", "m/s", float]
        self._columns[36] = ["dec", "Declination (J2000)", "dms", "U12"]
        self._columns[37] = ["ra", "Right ascension (J2000)", "hms", "U12"]

        self._ident = {
            "STAR": "pl_hostname",
            "NAME": "pl_name",
            "PER": "pl_orbper",
            "MASS": "pl_massj",
            "MSINI": "pl_msini",
            "R": "pl_radj",
            "DEPTH": "pl_trandep",
            "B": "pl_impact",
            "T14": "pl_trandur",
            "TT": "pl_tranmid",
            "T0": "pl_tperi",
            "A": "pl_orbsmax",
            "AR": "pl_orbsmaxr",
            "I": "pl_orbincl",
            "LAMBDA": "pl_missal",
            "OM": "pl_omega",
            "ECC": "pl_ecc",
            "GRAVITY": "pl_grav",
            "DENSITY": "pl_dens",
            "PLANETDISCMETH": "pl_dtype",
            "BINARY": "st_binary",
            "RSTAR": "st_rad",
            "DIST": "st_dist",
            "PAR": "st_par",
            "MSTAR": "st_mass",
            "TEFF": "st_teff",
            "LOGG": "st_logg",
            "VSINI": "st_vsini",
            "SHK": "st_acts",
            "RHK": "st_actc",
            "V": "st_vj",
            "FE": "st_fe",
            "GAMMA": "st_radv",
            "RHOSTAR": "st_dens",
            "DEC_STRING": "dec",
            "RA_STRING": "ra",
            "KOI": "KOI",
            "K": "K",
        }
        # Check whether data file exists
        self._fs = pp.PyAFS()
        if not skipUpdate:
            if self.needsUpdate() or (not self._fs.fileExists(self.dataFileName)):
                # Data needs update
                print("Downloading exoplanet data from explanets.org exoplanet archive")
                self._update(self._downloadData)
                print("Saved data to file: ", self.dataFileName, " in data directory,")
                print("  which has been configured as: ", self._fs.dpath)
                print("By default, the data will be downloaded anew every 7 days.")
                print("You can use the `changeDownloadCycle` to change this behavior.")
        self._readData()

    def downloadData(self):
        """
        Trigger download of data.
        """
        self._update(self._downloadData)

    def _readData(self):
        """
        Read the data from local file into numpy recarray.
        """
        r = csv.DictReader(
            self._fs.requestFile(self.dataFileName, "rt", gzip.open), delimiter=","
        )
        for nplanets, x in enumerate(r):
            pass
        r = csv.DictReader(
            self._fs.requestFile(self.dataFileName, "rt", gzip.open), delimiter=","
        )
        dtype = [
            (self._columns[x][0], self._columns[x][3])
            for x in range(len(self._columns))
        ]
        self.data = np.recarray((nplanets + 1,), dtype=dtype)

        # List of keys which shall be extracted
        desiredKeys = [x for x in six.iterkeys(self._ident)]

        # Assigned data
        for i, x in enumerate(r):
            for k in desiredKeys:
                v = x[k]
                if len(v) == 0:
                    v = None
                try:
                    self.data[self._ident[k]][i] = v
                except ValueError as ve:
                    self.data[self._ident[k]][i] = None
                    if self._verbose:
                        print(
                            "Assignment error encountered for: ",
                            self._ident[k],
                            ", i: ",
                            i,
                        )
                        print(x)
                        print(ve)
                        print()

    def availableColumns(self, verbose=True):
        """
        Shows a list of available data columns.

        Parameters
        ----------
        verbose : boolean, optional
            If True (default), prints information
            to screen.

        Returns
        -------
        columns : list of strings
            The names of the available data columns.
        """
        if self.data is None:
            return None
        if verbose:
            print("{0:12s}  {1:35s}  {2:5s}".format("Column", "Description", "Unit"))
            print("-" * 63)
            for v in six.itervalues(self._columns):
                print("{0:12s}  {1:35s}  {2:5s}".format(*v))
            print("-" * 63)

        return [self._columns[x][0] for x in smo.range(len(self._columns))]

    def selectByPlanetName(self, planetName, caseSensitive=False):
        """
        Get entry by planet name.

        Parameters
        ----------
        planetName : string
            The name of the planet (includes planet letter,
            e.g., "corot-2 b"
        caseSensitive : boolean, optional
            If False (default), the search will be case-insensitive.

        Returns
        -------
        Data entry : dictionary
            A dictionary with a key for every data column holding
            the associated value from the data table.
        """
        try:
            plnames = [name.decode("utf8") for name in self.data.pl_name]
        except AttributeError:
            plnames = [name for name in self.data.pl_name]
        r = pyaC.fuzzyMatch(
            planetName, plnames, caseSensitive=caseSensitive, raises=True
        )
        result = {}
        for c in six.itervalues(self._columns):
            result[c[0]] = self.data[c[0]][r["index"]]
        return result

    def getAllData(self):
        """
        Get all available data.

        Returns
        -------
        Data : numpy recarray
            All data stored in the table as a numpy recarray.
        """
        return self.data.copy()




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Convenience function for fitting
        # ==============================
        
        # In[1]:
        
        
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones_like
        import matplotlib.pylab as plt
        import scipy.optimize as sco
        # ... and now the funcFit2 package
        from PyAstronomy import funcFit2 as fuf2
        import matplotlib
        matplotlib.use('Agg')
        
        # For reproducability
        random.seed(1234)
        
        # Create some mock data
        # Choose parameters...
        gPar = {"A":-5.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # ...and calculate profile...
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2*pi*gPar["sig"]**2) \
            * exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # ...add some noise...
        y += random.normal(0.0, 0.01, x.size)
        # ...and save the error bars
        yerr = ones_like(x)*0.01
        
        plt.errorbar(x, y, yerr=yerr, fmt='b+', label="Data")
        plt.legend()
        plt.show()
        
        # In[2]:
        
        
        # Create a fitting object representing a Gaussian and set guess parameters.
        gf = fuf2.GaussFit()
        
        # Set guess values for the parameters
        gf["A"] = -10.0
        gf["sig"] = 15.77
        gf["off"] = 0.96
        gf["mu"] = 7.5
        
        # Thaw parameters to be treated as variable during the fit
        gf.thaw(["A", "sig", "off", "mu"])
        
        # Save starting parameters
        pstart = gf.parameters()
        
        # Use convenience function
        # ---------------------------------------
        
        # In[3]:
        
        
        # Use a convenience function to carry out optimization
        # using various algorithms from scipy
        
        # Let them start from the same starting point
        gf.assignValues(pstart)
        fr1 = fuf2.fitfmin(gf, gf.chisqr, x, y, yerr=yerr, disp=0)
        # Save resulting model
        model1 = gf.evaluate(x)
        
        gf.assignValues(pstart)
        fr2 = fuf2.fitfmin_powell(gf, gf.chisqr, x, y, yerr=yerr, disp=0)
        model2 = gf.evaluate(x)
        
        gf.assignValues(pstart)
        fr3 = fuf2.fitfmin_l_bfgs_b(gf, gf.chisqr, x, y, yerr=yerr)
        model3 = gf.evaluate(x)
        
        # In[4]:
        
        
        plt.errorbar(x, y, yerr=yerr, fmt='b+', label="Data")
        plt.plot(x, model1, 'c-', label="fmin")
        plt.errorbar(x, y+0.05, yerr=yerr, fmt='b+')
        plt.plot(x, model2+0.05, 'r-', label="fminpowell")
        plt.errorbar(x, y+0.1, yerr=yerr, fmt='b+')
        plt.plot(x, model3+0.1, 'g-', label="l_bfgs_b")
        plt.legend()
        plt.show()
        
        # In[5]:
        
        
        
        



from __future__ import print_function, division
import scipy.interpolate as sci
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE


def dopplerShift(wvl, flux, v, edgeHandling=None, fillValue=None, vlim=0.05, err=None, wvlOrdered=False):
    """
    Doppler shift a given spectrum.

    An algorithm to apply a Doppler shift
    to a spectrum. The idea here is to obtain a shifted spectrum
    without loosing the wavelength axis. Therefore,
    this function, first, calculates
    the shifted wavelength axis and, second, obtains
    the new, shifted flux array at the old, unshifted
    wavelength points by linearly interpolating. No
    relativistic effects are considered.

    Due to the shift, some bins at the edge of the
    spectrum cannot be interpolated, because they
    are outside the given input range. The default
    behavior of this function is to return numpy.NAN
    values at those points. One can, however, specify
    the `edgeHandling` parameter to choose a different
    handling of these points.

    If "firstlast" is specified for `edgeHandling`,
    the out-of-range points at the red or blue edge
    of the spectrum will be filled using the first
    (at the blue edge) and last (at the red edge) valid
    point in the shifted, i.e., the interpolated, spectrum.

    If "fillValue" is chosen for edge handling,
    the points under consideration will be filled with
    the value given through the `fillValue` keyword.

    .. warning:: Shifting a spectrum using linear
                interpolation has an effect on the
                noise of the spectrum. No treatment
                of such effects is implemented in this
                function.

    Parameters
    ----------
    wvl : array
        Input wavelengths in A.
    flux : array
        Input flux.
    v : float
        Doppler shift in km/s
    edgeHandling : string, {"fillValue", "firstlast"}, optional
        The method used to handle the edges of the
        output spectrum. If not specified, NaN will be used as
        fill value.
    fillValue : float, optional
        If the "fillValue" is specified as edge handling method,
        the value used to fill the edges of the output spectrum.
    vlim : float, optional
        Maximal fraction of the speed of light allowed for Doppler
        shift, v. Default is 0.05.
    err : array, optional
        The uncertainties of the data points in the spectrum. If given,
        updated uncertainties are obtained by linear interpolation of the
        uncertainties.
    wvlOrdered : bool, optional
        If True, the check of the ordering of the wavelength axis will be
        skipped. Default is False.

    Returns
    -------
    nflux : array
        The shifted flux array at the *old* input locations.
    wlprime : array
        The shifted wavelength axis.
    err : array, optional
        Uncertainties of the shifted spectrum.
    """
    # Order check
    if not wvlOrdered:
        if np.any(np.diff(wvl) < 0.0):
            raise(PE.PyAValError("Wavelength axis must be sorted in ascending order.", \
                                 solution="Use sorted axis.")) 
        
    # Speed of light [km/s]
    cvel = 299792.458
    
    if np.abs(v) > vlim*cvel:
        raise(PE.PyAValError("Specified velocity of % g km/s exceeds %g %% of the speed of light. No relativistic effects are considered in this implementation." % (v, vlim*100.), \
                             where="dopplerShift", \
                             solution="Increase 'vlim' if you wish to suppress this error."))
    
    # Shifted wavelength axis
    wlprime = wvl * (1.0 + v / cvel)

    # Overlap check
    if (wlprime[0] >= wvl[-1]) or (wlprime[-1] <= wvl[0]):
        raise(PE.PyAValError("The shifted wavelength axis shows no overlap with the input axis. The velocity shift of %g km/s is too large." % v, \
                             where="dopplerShift", \
                             solution=["Use smaller shifts", "Please consider another implementation. Also note that the treatment here is not relativistic."]))

    fv = np.nan
    if edgeHandling == "fillValue":
        if fillValue is None:
            raise(PE.PyAValError("Fill value not specified", where="pyasl.dopplerShift",
                                 solution="If you request 'fillValue' as edge handling method, you need to specify the 'fillValue' keyword."))
        fv = fillValue

    nflux = sci.interp1d(wlprime, flux, bounds_error=False, fill_value=fv)(wvl)
    if err is not None:
        nerr = sci.interp1d(wlprime, err, bounds_error=False, fill_value=fv)(wvl)

    if edgeHandling == "firstlast":
        # Not is-NaN
        nin = ~np.isnan(nflux)
        if not nin[0]:
            # First element is invalid (NaN)
            # Find index of first valid (not NaN) element
            fvindex = np.argmax(nin)
            # Replace leading elements
            nflux[0:fvindex] = nflux[fvindex]
            if err is not None:
                nerr[0:fvindex] = nerr[fvindex]
        if not nin[-1]:
            # Last element is invalid
            # Index of last valid element
            lvindex = -np.argmax(nin[::-1])-1
            # Replace trailing elements
            nflux[lvindex+1:] = nflux[lvindex]
            if err is not None:
                nerr[lvindex+1:] = nerr[lvindex]

    if err is None:
        return nflux, wlprime
    else:
        return nflux, wlprime, nerr



from __future__ import print_function, division
import numpy
from PyAstronomy.funcFit.utils import ic
from PyAstronomy.pyaC import pyaErrors as PE
import scipy.misc as sm


def bayesFactors(model1, model2):
    r"""
    Computes the Bayes factor for two competing models.

    The computation is based on Newton & Raftery 1994 (Eq. 13).

    Parameters
    ----------
    model1 : PyAstronomy.funcFit.TraceAnalysis instance
        TraceAnalysis for the first model.
    model2 : PyAstronomy.funcFit.TraceAnalysis instance
        TraceAnalysis for the second model.

    Returns
    -------
    BF : float
        The Bayes factor BF (neither log10(BF) nor ln(BF)).
        Note that a small value means that the first model
        is favored, i.e., BF=p(M2)/p(M1).

    """
    logp1 = model1["deviance"] / (-2.)
    logp2 = model2["deviance"] / (-2.)

    # Given a number of numbers x1, x2, ... whose logarithms are given by l_i=ln(x_i) etc.,
    # logsum calculates: ln(sum_i exp(l_i)) = ln(sum_i x_i)
    bf = numpy.exp(sm.logsumexp(-logp1) - numpy.log(len(logp1))
                   - (sm.logsumexp(-logp2) - numpy.log(len(logp2))))
    return bf



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import AtomicNo
import os
from PyAstronomy.pyasl import _ic
import six

class FirstIonizationPot:
  """
    First ionization potentials of individual elements.
    
    The data have been obtained from NIST:
    http://physics.nist.gov/PhysRefData/ASD/ionEnergy.html
  """

  def _decomposeFIP(self, s):
    """
      Decompose value string.
      
      Parameters
      ----------
      s : string
          The specification of the value. Values in
          round parentheses denote theoretical values,
          values in brackets are interpolated.
      
      Returns
      -------
      Value : float
          The first ionization energy [eV].
      Error: float
          The error of the first ionization energy [eV].
      isTheoretical : boolean
          True if value is theoretical.
      isInterpolated : boolean
          True if value has been interpolated.
    """
    isTheoretical = False
    isInterpolated = False
    
    s = s.strip()
    if s.startswith("("):
      isTheoretical = True
      s = s[1:-1]
    if s.startswith("["):
      isInterpolated = True
      s = s[1:-1]
    # Isolate error
    if s.find(")") != -1 :
      # There is an error given
      s = s.rstrip(")")
      r = s.split("(")
      error = float(r[1])
      # Determine number of significant digits
      ndig = len(r[0].split(".")[1])
      error *= 10.0**(-ndig)
    else:
      # No error was given
      error = None
      r = [s]
    val = float(r[0])
    return val, error, isTheoretical, isInterpolated

  def getFIP(self, atom):
    """
      Get the first ionization energy.
      
      Parameters
      ----------
      atom : string or integer
          Either a string specifying the elemental symbol (e.g., 'H') or
          an integer specifying the atomic number.
      
      Returns
      -------
      FIP : float
          First ionization potential [eV].
      FIP error : float
          Error of the FIP [eV]. May also be None
          if no error is available.
    """
    if isinstance(atom, six.string_types):
      an = self._an.getAtomicNo(atom)
    elif isinstance(atom, int):
      an = atom
    else:
      raise(PE.PyAValError("'atom' must be an integer specifying the atomic number are a string holding the elemental symbol.", \
                           where="FirstIonizationPot::getFIP"))
    return self._fip[an][0], self._fip[an][1]
    

  def __init__(self):
    self._fip = {}
    # Build FIP-data dictionary. Uses elemental number as key
    # and, for each, holds a tuple composed of value, error, a flag
    # describing whether the value is theoretical, and a flag specifying
    # whether the value has been interpolated (see NIST link for further)
    # details.
    import PyAstronomy as _PyA
    fn = os.path.join(os.path.dirname(_PyA.__file__), "pyasl", "resBased", "fip.dat")
    for l in open(fn):
      if l.startswith("#"):
        continue
      s = l.split("|")
      self._fip[int(s[0])] = self._decomposeFIP(s[8])
    # Convert between atomic number and elemental symbol
    self._an = AtomicNo()


def plotFIP():
  """
    Show a plot of first ionization energy vs. atomic number. 
  """
  if not _ic.check["matplotlib"]:
    raise(PE.PyARequiredImport("Could not import matplotlib."))
  
  fip = FirstIonizationPot()
  z = range(1,31,1)
  fips = []
  for an in z:
    fips.append(fip.getFIP(an)[0])
  
  import matplotlib.pylab as plt
  plt.xlabel("Atomic number")
  plt.ylabel("First ionization energy [eV]")
  plt.plot(z, fips, 'bp-')
  plt.show()
  
  
  